from typing import Callable, Dict, List, Optional, Union

import solara
from solara.components.component_vue import component_vue


@component_vue("menu.vue")
def MenuWidget(
    activator: List[solara.Element],
    show_menu: bool,
    on_show_menu: Optional[Callable] = None,
    close_on_content_click: bool = True,
    children: List[solara.Element] = [],
    style: Optional[str] = None,
    context: bool = False,
    use_absolute: bool = True,
    use_activator_width: bool = True,
):
    pass


@solara.component
def ClickMenu(
    activator: Union[solara.Element, List[solara.Element]],
    open_value: Union[solara.Reactive[bool], bool] = False,
    on_open_value: Optional[Callable] = None,
    children: List[solara.Element] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """
    Show a pop-up menu by clicking on the `activator` element. The menu appears at the cursor position.

    ```solara
    import solara


    @solara.component
    def Page():
        image_url = "/static/public/beach.jpeg"
        image = solara.Image(image=image_url)

        with solara.lab.ClickMenu(activator=image):
            with solara.Column(gap="0px"):
                [solara.Button(f"Click me {i}!", text=True) for i in range(5)]

    ```


    ## Arguments

    * activator: Clicking on this element will open the menu. Accepts either a `solara.Element`, or a list of elements.
    * open_value: Controls and communicates the state of the menu. If True, the menu is open. If False, the menu is closed.
    * on_open_value: Function to call when the menu is opened or closed.
    * menu_contents: List of Elements to be contained in the menu.
    * style: CSS style to apply. Applied directly onto the `v-menu` component.
    """
    open_reactive = solara.use_reactive(open_value, on_open_value)
    del open_value

    style_flat = solara.util._flatten_style(style)

    if not isinstance(activator, list):
        activator = [activator]

    return MenuWidget(
        activator=activator,
        children=children,
        show_menu=open_reactive.value,
        on_show_menu=open_reactive.set,
        style=style_flat,
    )


@solara.component
def ContextMenu(
    activator: Union[solara.Element, List[solara.Element]],
    open_value: Union[solara.Reactive[bool], bool] = False,
    on_open_value: Optional[Callable] = None,
    children: List[solara.Element] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """
    Show a context menu by triggering the contextmenu event on the `activator` element. The menu appears at the cursor position.

    A contextmenu event is typically triggered by clicking the right mouse button, or by pressing the context menu key.

    ```solara
    import solara


    @solara.component
    def Page():
        image_url = "/static/public/beach.jpeg"
        image = solara.Image(image=image_url)

        with solara.lab.ContextMenu(activator=image):
            with solara.Column(gap="0px"):
                [solara.Button(f"Click me {i}!", text=True) for i in range(5)]

    ```

    ## Arguments

    * activator: Clicking on this element will open the menu. Accepts either a `solara.Element`, or a list of elements.
    * open_value: Controls and communicates the state of the menu. If True, the menu is open. If False, the menu is closed.
    * on_open_value: Function to call when the menu is opened or closed.
    * children: List of Elements to be contained in the menu
    * style: CSS style to apply. Applied directly onto the `v-menu` component.
    """
    open_reactive = solara.use_reactive(open_value, on_open_value)
    del open_value
    style_flat = solara.util._flatten_style(style)

    if not isinstance(activator, list):
        activator = [activator]

    return MenuWidget(
        activator=activator,
        children=children,
        show_menu=open_reactive.value,
        on_show_menu=open_reactive.set,
        style=style_flat,
        context=True,
    )


@solara.component
def Menu(
    activator: Union[solara.Element, List[solara.Element]],
    open_value: Union[solara.Reactive[bool], bool] = False,
    on_open_value: Optional[Callable] = None,
    close_on_content_click: bool = True,
    children: List[solara.Element] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
    use_activator_width: bool = True,
):
    """
    Show a pop-up menu by clicking on the `activator` element. The menu appears below the `activator` element.

    ```solara
    import solara


    @solara.component
    def Page():
        btn = solara.Button("Show suboptions")

        with solara.lab.Menu(activator=btn):
            with solara.Column(gap="0px"):
                [solara.Button(f"Click me {str(i)}!", text=True) for i in range(5)]

    ```

    ## Arguments

    * activator: Clicking on this element will open the menu. Accepts either a `solara.Element`, or a list of elements.
    * open_value: Controls and communicates the state of the menu. If True, the menu is open. If False, the menu is closed.
    * on_open_value: Function to call when the menu is opened or closed.
    * children: List of Elements to be contained in the menu
    * style: CSS style to apply. Applied directly onto the `v-menu` component.
    * use_activator_width: If True, the menu will have a minimum width equal to the activator element.
      If False, the menu width will be determined by the content.
    """
    open_reactive = solara.use_reactive(open_value, on_open_value)
    del open_value

    style_flat = solara.util._flatten_style(style)

    if not isinstance(activator, list):
        activator = [activator]

    return MenuWidget(
        activator=activator,
        children=children,
        show_menu=open_reactive.value,
        on_show_menu=open_reactive.set,
        close_on_content_click=close_on_content_click,
        style=style_flat,
        use_absolute=False,
        use_activator_width=use_activator_width,
    )



import solara
from solara.alias import rv


@solara.component
def ColorCard(title, color):
    with rv.Card(style_=f"background-color: {color}; width: 100%; height: 100%") as main:
        rv.CardTitle(children=[title])
    return main



"""# HTML"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    html = """
<h1>Custom html</h1>
<ul>
    <li>Item 1
    <li>Item 2
</ul>
"""
    solara.HTML(tag="div", unsafe_innerHTML=html)


__doc__ += apidoc(solara.HTML.f)  # type: ignore



"""# Deploy model demo

This show off a combination of [solara.lab.Menu](/documentation/components/lab/menu) and
    [solara.lab.ConfirmationDialog](/documentation/components/lab/confirmation_dialog)
and [solara.use_thread](/documentation/api/hooks/use_thread) to demonstrate how to tune, fit and deploy a model.
The actually deployment is not real, but simulated by a thread that returns a boolean value on success and sleep for a while to
similate the deployment taking time.

 * Right click on the plot to either fit or reset the model parameters.
 * Clicking on the button shows a custom menu to deploy to several environments.
 * A confirmation dialog is used before executing the deployment.
 * If the deployment fails, a confirmation dialog is shown to retry the deployment.
"""

import time

import numpy as np
import pandas as pd
import plotly.express as px
import solara
import solara.lab

slope = solara.reactive(1.5)
intercept = solara.reactive(0.5)

x = np.arange(0, 100, 2) / 10
# fake data with noise
y_data = (slope.value * 0.8) * x + (intercept.value + 0.4) + np.random.normal(0, 0.5, len(x))


@solara.component
def Page():
    deploy_environment_request = solara.use_reactive("")
    deploy_environment = solara.use_reactive("")
    fake_failed = solara.use_reactive(False)

    def deploy_model():
        if deploy_environment.value == "":
            return

        # 'dummy' deployment
        time.sleep(1.5)

        # fake a failure the first time
        if not fake_failed.value:
            fake_failed.value = True
            return False
        else:
            return True

    deploy_state: solara.Result[bool] = solara.use_thread(deploy_model, dependencies=[deploy_environment.value])

    x_model = x
    y_model = (slope.value) * x_model + (intercept.value)
    df1 = pd.DataFrame({"x": x, "y": y_data})
    df2 = pd.DataFrame({"x": x_model, "y": y_model})
    df = pd.concat([df1, df2], keys=["data", "model"], names=["type"]).reset_index()
    fig = px.line(df, x="x", y="y", color="type")
    with solara.Card("My fancy model"):
        fig_element = solara.FigurePlotly(fig)

        def reset():
            slope.set(1.5)
            intercept.set(0.5)

        def fit():
            slope.set(1.5 * 0.8)
            intercept.set(0.9)

        with solara.lab.ContextMenu(activator=fig_element):
            solara.Button("Reset", text=True, icon_name="mdi-refresh", on_click=reset)
            solara.Button("Fit", text=True, icon_name="mdi-brain", on_click=fit)

        solara.FloatSlider("Slope", value=slope, min=-2, max=5, step=0.1)
        solara.FloatSlider("Intercept", value=intercept, min=-10, max=10, step=0.1)

        solara.lab.ConfirmationDialog(
            deploy_environment_request.value != "",
            title="Confirm deployment",
            content=solara.Markdown(f"Are you sure you want to deploy to **{deploy_environment_request.value}**?"),
            ok=f"Deploy to {deploy_environment_request.value}",
            on_ok=lambda: deploy_environment.set(deploy_environment_request.value),
            on_close=lambda: deploy_environment_request.set(""),
        )

        solara.lab.ConfirmationDialog(
            deploy_state.value is False and deploy_state.state == solara.ResultState.FINISHED,
            title="Deployment failed",
            on_close=lambda: None,
            on_ok=lambda: deploy_state.retry(),
            content=solara.Error(solara.Markdown(f"Deployment to **{deploy_environment.value}** failed. Do you want to try again?")),
            ok="Retry deployment",
            cancel="Ignore the failure",
        )

        def request_deploy(environment):
            deploy_environment.value = ""
            deploy_environment_request.set(environment)

        deploying = bool(deploy_environment.value and deploy_state.state == solara.ResultState.RUNNING)
        deploy_button = solara.Button("Deploy model", color="primary", icon_name="mdi-rocket", loading=deploying)
        with solara.lab.Menu(activator=deploy_button):
            with solara.Column(gap="0px"):
                solara.Button("Deploy to testing environment", text=True, icon_name="mdi-bug", on_click=lambda: request_deploy("testing"))
                solara.Button("Deploy to staging environment", text=True, icon_name="mdi-test-tube", on_click=lambda: request_deploy("staging"))
                solara.Button(
                    "Deploy to production",
                    text=True,
                    icon_name="mdi-rocket",
                    style={"justify-content": "left"},  # make the icon appear on the left
                    on_click=lambda: request_deploy("production"),
                )
        if deploy_state.value is True and deploy_state.state == solara.ResultState.FINISHED:
            # add some margin in the y direction using my-4
            solara.Success(solara.Markdown(f"Model deployed successfully to **{deploy_environment.value}**!"), classes=["my-4"])



# isort: skip_file
from .button import Button  # noqa: F401
from .style import Style  # noqa: F401 F403
from .misc import *  # noqa: F401 F403
from .alert import Warning, Info, Error, Success  # noqa: F401 F403
from .checkbox import Checkbox  # noqa: F401 F403
from .cross_filter import (  # noqa: F401
    CrossFilterDataFrame,
    CrossFilterReport,
    CrossFilterSelect,
    CrossFilterSlider,
)
from .datatable import DataTable, DataFrame  # noqa: F401 F403
from .details import Details  # noqa: F401 F403
from .file_browser import FileBrowser  # noqa: F401 F403
from .image import Image  # noqa: F401 F403
from .markdown import Markdown, MarkdownIt  # noqa: F401 F403
from .slider import (  # noqa: F401 F403
    DateSlider,  # noqa: F401 F403
    FloatSlider,  # noqa: F401 F403
    IntSlider,  # noqa: F401 F403
    ValueSlider,  # noqa: F401 F403
    SliderDate,  # noqa: F401 F403
    SliderFloat,  # noqa: F401 F403
    SliderInt,  # noqa: F401 F403
    SliderValue,  # noqa: F401 F403
    SliderRangeInt,  # noqa: F401 F403
    SliderRangeFloat,  # noqa: F401 F403
)  # noqa: F401 F403
from .sql_code import SqlCode  # noqa: F401 F403
from .togglebuttons import (  # noqa: F401 F403
    ToggleButtonsMultiple,
    ToggleButtonsSingle,
)
from .input import InputText, InputFloat, InputInt  # noqa: F401 F403
from .input_text_area import InputTextArea  # noqa: F401 F403
from .pivot_table import PivotTableView, PivotTable, PivotTableCard  # noqa: F401 F403
from .head import Head  # noqa: F401 F403
from .title import Title  # noqa: F401 F403
from .link import Link  # noqa: F401 F403
from .applayout import AppLayout, Sidebar, AppBar, AppBarTitle  # noqa: F401 F403
from .tab_navigation import TabNavigation  # noqa: F401 F403
from .markdown_editor import MarkdownEditor  # noqa: F401 F403
from .select import Select, SelectMultiple  # noqa: F401 F403
from .matplotlib import FigureMatplotlib  # noqa: F401 F403
from .echarts import FigureEcharts  # noqa: F401 F403
from .figure_altair import FigureAltair, AltairChart  # noqa: F401 F403
from .meta import Meta  # noqa: F401 F403
from .columns import Columns, ColumnsResponsive  # noqa: F401 F403
from .file_drop import FileDrop, FileDropMultiple  # noqa: F401 F403
from .file_download import FileDownload  # noqa: F401 F403
from .tooltip import Tooltip  # noqa: F401 F403
from .card import Card, CardActions  # noqa: F401 F403
from .spinner import SpinnerSolara  # noqa: F401 F403
from .switch import Switch  # noqa: F401 F403
from .progress import ProgressLinear  # noqa: F401 F403
from .component_vue import _component_vue, component_vue  # noqa: F401 F403
import reacton.core

try:
    from reacton import Fragment as Fragment  # type: ignore
except ImportError:
    pass

import logging
from ..settings import main

_container = None

if main.default_container in globals():
    _container = globals()[main.default_container]
else:
    logger = logging.getLogger("solara.components")
    logger.warning(f"Default container {main.default_container} not found in solara.components. Defaulting to Column.")

# TODO: When Solara 2.0 releases Column should be replaced with Fragment
reacton.core._default_container = _container or Column  # noqa: F405



"""
# computed

"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "computed"


Page = NoPage


__doc__ += apidoc(solara.lab.computed)  # type: ignore



# tests pages in the api docs
import playwright.sync_api
from playwright.sync_api import expect

# def test_api_sqlcode(page: playwright.sync_api.Page, solara_server, solara_app):
#     # with screenshot_on_error(page, 'tmp/test_docs_basics.png'):
#     with solara_app("solara.website.pages"):
#         page.goto(solara_server.base_url + "/api")
#         page.locator("text=Sql Code").click()
#         page.locator('button:has-text("Execute")').wait_for()


# def test_api_grid_draggable(page: playwright.sync_api.Page, solara_server, solara_app):
#     # with screenshot_on_error(page, 'tmp/test_docs_basics.png'):
#     with solara_app("solara..website.pages"):
#         page.goto(solara_server.base_url + "/api")
#         page.locator("text=GridDraggable").click()
#         page.locator('button:has-text("Reset to initial layout")').wait_for()


def test_api_markdown_editor(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/components/")
        page_session.locator(".v-card >> text=Markdown Editor").first.click()
        page_session.locator('h1:has-text("Large heading")').wait_for()


def test_api_file_browser(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/components/")
        page_session.locator(".v-card >> text=File Browser").first.click()
        page_session.locator("text=You are in directory").wait_for()


def test_api_matplotlib(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/components/viz/matplotlib")
        page_session.locator("text=Arguments").first.wait_for()


def test_api_style(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/components/advanced/style")
        page_session.locator("text=Add a custom piece of CSS").first.wait_for()
        expect(page_session.locator(".mybutton")).to_have_css("color", "rgb(76, 175, 80)")

        # disable css
        page_session.locator('_vue=v-checkbox[label="Use CSS"]').click()
        expect(page_session.locator(".mybutton")).to_have_css("color", "rgba(0, 0, 0, 0.87)")

        # enable css again
        page_session.locator('_vue=v-checkbox[label="Use CSS"]').click()
        expect(page_session.locator(".mybutton")).to_have_css("color", "rgb(76, 175, 80)")


def test_api_cross_filter_select(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/api/cross_filter/cross_filter_select")
        page_session.locator("text=244").wait_for()
        select_sex = page_session.locator('_vue=v-autocomplete[label="Select values in sex having values:"]')
        select_sex.click()
        page_session.locator("text=Female").click()
        page_session.locator("text=87 / 244").wait_for()


def test_dataframe(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url + "/documentation/components/data/dataframe")
        datatable_value_counts = page_session.locator(".solara-data-table").nth(1)
        datatable_value_counts.wait_for()

        # open the species hover menu
        datatable_value_counts.locator("th >> text=species_id >> _vue=v-icon").hover()
        page_session.locator("text=Value counts for species_id").wait_for()
        page_session.locator("text=Name: count").wait_for()

        # open the petal_width hover menu
        datatable_value_counts.locator("th >> text=petal_width >> _vue=v-icon").hover()
        page_session.locator("text=Name species").wait_for(state="detached")
        page_session.locator("text=Value counts for petal_width").wait_for()
        page_session.locator("text=Name: count").wait_for()



"""# Tooltip"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

Page = NoPage
title = "Tooltip"


__doc__ += apidoc(solara.Tooltip.f)  # type: ignore



import platform
import unittest.mock
from pathlib import Path
from typing import Optional, cast

import pytest

import solara
import solara.components.file_browser

HERE = Path(__file__)


def test_file_browser_callback_no_select():
    on_directory_change = unittest.mock.MagicMock()
    on_file_open = unittest.mock.MagicMock()
    on_file_name = unittest.mock.MagicMock()  # backwards compat test
    on_path_select = unittest.mock.MagicMock()

    @solara.component
    def Test():
        return solara.FileBrowser(
            HERE.parent, on_path_select=on_path_select, on_directory_change=on_directory_change, on_file_open=on_file_open, on_file_name=on_file_name
        )

    div, rc = solara.render_fixed(Test(), handle_error=False)
    on_directory_change.assert_not_called()
    on_file_open.assert_not_called()
    on_path_select.assert_not_called()

    list: solara.components.file_browser.FileListWidget = div.children[1]
    with pytest.raises(NameError, match=".*foo.*"):
        list.test_click("foo")
    # select ..
    list.test_click("..")
    on_directory_change.assert_called_once()
    on_file_open.assert_not_called()
    on_path_select.assert_not_called()
    assert "conftest.py" in list

    # select conftest.py
    list.test_click("conftest.py")
    conftest_path = Path(HERE).parent.parent / "conftest.py"
    on_file_open.assert_called_with(conftest_path)
    on_file_name.assert_called_with(str(conftest_path))
    on_path_select.assert_not_called()
    assert list.clicked is not None
    assert list.clicked["name"] == "conftest.py"

    list.clicked = None  # type: ignore
    on_file_open.assert_called_once()
    on_path_select.assert_not_called()

    list.test_click("unit")
    assert list.clicked is None


def test_file_browser_callback_can_select():
    on_directory_change = unittest.mock.MagicMock()
    on_file_open = unittest.mock.MagicMock()
    on_path_select = unittest.mock.MagicMock()

    @solara.component
    def Test():
        return solara.FileBrowser(
            HERE.parent, on_path_select=on_path_select, on_directory_change=on_directory_change, on_file_open=on_file_open, can_select=True
        )

    div, rc = solara.render_fixed(Test())
    on_directory_change.assert_not_called()
    on_file_open.assert_not_called()
    on_path_select.assert_not_called()

    list: solara.components.file_browser.FileListWidget = div.children[1]
    assert "file_browser_test.py" in list

    # select ..
    list.test_click("..")
    on_directory_change.assert_not_called()
    on_file_open.assert_not_called()
    on_path_select.assert_called_with(HERE.parent.parent)

    # change to ..
    list.test_click("..", double_click=True)
    on_directory_change.assert_called_once()
    on_file_open.assert_not_called()
    on_path_select.assert_called_with(None)
    assert "conftest.py" in list

    list.test_click("conftest.py")
    conftest_path = Path(HERE).parent.parent / "conftest.py"
    on_path_select.assert_called_with(conftest_path)

    # enter the current directory again
    list.test_click("unit", double_click=True)
    # we shouldn't have a file with the same name selected any more
    on_path_select.assert_called_with(None)
    assert list.double_clicked is None

    # open conftest
    list.test_click("conftest.py", double_click=True)
    conftest_path = Path(HERE).parent / "conftest.py"
    on_file_open.assert_called_with(conftest_path)

    list.clicked = None
    on_path_select.assert_called_with(None)

    # go up
    list.test_click("..", double_click=True)
    assert "conftest.py" in list

    # go up again
    list.test_click("..", double_click=True)
    assert "conftest.py" not in list


def test_file_browser_scroll_pos():
    @solara.component
    def Test():
        return solara.FileBrowser(HERE.parent)

    div, rc = solara.render_fixed(Test())

    list: solara.components.file_browser.FileListWidget = div.children[1]
    assert "file_browser_test.py" in list

    # select ..
    list.test_click("..")
    assert "unit" in list
    list.scroll_pos = 10

    # go to unit
    list.test_click("unit")

    # and back to ..
    list.test_click("..")
    assert list.scroll_pos == 10


@pytest.mark.skipif(platform.system() == "Windows", reason="Windows doesn't support chmod")
def test_file_browser_no_access(tmpdir: Path):
    on_directory_change = unittest.mock.MagicMock()
    on_file_open = unittest.mock.MagicMock()
    on_path_select = unittest.mock.MagicMock()

    path_no_read = tmpdir / "no_read"
    path_no_read.mkdir()
    mode = path_no_read.stat().mode  # type: ignore
    path_no_read.chmod(000)

    @solara.component
    def Test():
        return solara.FileBrowser(
            Path(tmpdir), on_path_select=on_path_select, on_directory_change=on_directory_change, on_file_open=on_file_open, can_select=True
        )

    try:
        div, rc = solara.render_fixed(Test(), handle_error=False)

        list: solara.components.file_browser.FileListWidget = div.children[1]
        # select is ok
        list.test_click("no_read")

        # enter is not
        on_path_select.assert_called_with(path_no_read)
        list.test_click("no_read", double_click=True)
        on_directory_change.assert_not_called()
    finally:
        path_no_read.chmod(mode)


def test_file_browser_filter():
    def directory_filter(path: Path) -> bool:
        return path.is_dir() and not path.name.startswith("_")

    @solara.component
    def Test():
        return solara.FileBrowser(HERE.parent.parent, filter=directory_filter)

    div, rc = solara.render_fixed(Test())

    list: solara.components.file_browser.FileListWidget = div.children[1]
    items = list.files
    names = {k["name"] for k in items}
    assert names.issuperset({"unit", "ui", "docs", "integration", "pyinstaller", ".."})


def test_file_browser_test_change_directory():
    div, rc = solara.render_fixed(solara.FileBrowser(HERE.parent))
    list: solara.components.file_browser.FileListWidget = div.children[1]
    assert "file_browser_test.py" in list
    rc.render(solara.FileBrowser(HERE.parent.parent))
    assert "file_browser_test.py" not in list


def test_file_browser_control_directory():
    import solara

    def directory_filter(path: Path) -> bool:
        return path.is_dir() and not path.name.startswith("_")

    BASE_PATH = HERE.parent.parent

    @solara.component
    def Page():
        def set_directory(path: Path) -> None:
            directory.value = path if str(path).startswith(str(BASE_PATH)) else BASE_PATH

        directory = solara.use_reactive(BASE_PATH, set_directory)
        solara.FileBrowser(directory, filter=directory_filter)

    _, rc = solara.render(Page(), handle_error=False)
    file_list = rc.find(solara.components.file_browser.FileListWidget).widget
    mock = unittest.mock.MagicMock()
    file_list.observe(mock, "files")
    items = file_list.files
    names = {k["name"] for k in items}
    assert names.issuperset({"unit", "ui", "docs", "integration", "pyinstaller", ".."})
    file_list.test_click("..")
    assert mock.call_count == 0
    file_list.test_click("integration")
    items = file_list.files
    names = {k["name"] for k in items}
    assert names != {"unit", "ui", "docs", "integration", "pyinstaller", ".."}
    assert mock.call_count == 1


def test_file_browser_relative_path():
    @solara.component
    def Test():
        return solara.FileBrowser(".")

    div, rc = solara.render_fixed(Test())
    list: solara.components.file_browser.FileListWidget = div.children[1]
    files = {k["name"] for k in list.files}
    list.test_click("..")
    files_parent = {k["name"] for k in list.files}
    assert files_parent != files


def test_file_browser_programmatic_select():
    # using a reactive value to select a file
    selected = solara.reactive(cast(Optional[Path], None))
    current_dir = solara.reactive(HERE.parent)

    @solara.component
    def Test():
        return solara.FileBrowser(HERE.parent, selected=selected, can_select=True)

    div, rc = solara.render_fixed(Test(), handle_error=False)
    list: solara.components.file_browser.FileListWidget = div.children[1]
    files = list.files.copy()
    assert list.clicked is None
    selected.value = HERE.parent / "file_browser_test.py"
    assert list.clicked is not None
    assert list.clicked["name"] == "file_browser_test.py"
    list.test_click("..", double_click=True)
    assert list.files != files
    assert selected.value is None

    selected.value = None
    rc.close()

    # passing selected as a value (non reactive)
    @solara.component
    def Test2():
        return solara.FileBrowser(HERE.parent, selected=selected.value, on_path_select=selected.set, can_select=True)

    div, rc = solara.render_fixed(Test2(), handle_error=False)
    list = div.children[1]
    files = list.files.copy()
    assert list.clicked is None
    selected.value = HERE.parent / "file_browser_test.py"
    assert list.clicked is not None
    assert list.clicked["name"] == "file_browser_test.py"
    list.test_click("..", double_click=True)
    assert list.files != files
    assert selected.value is None
    rc.close()

    @solara.component
    def Test3():
        return solara.FileBrowser(current_dir, selected=selected.value, on_path_select=selected.set, can_select=True)

    div, rc = solara.render_fixed(Test3(), handle_error=False)
    assert current_dir.value == HERE.parent
    list.test_click("..", double_click=False)
    assert current_dir.value == HERE.parent
    # this will trigger the sync_directory_from_selected
    selected.value = current_dir.value / ".." / "unit" / ".."
    assert current_dir.value == HERE.parent
    rc.close()



import pathlib
import sys
from typing import Optional, cast

import pandas as pd

import solara
import solara.express as solara_px  # similar to plotly express, but comes with cross filters
import solara.lab
from solara.components.columns import Columns
from solara.components.file_drop import FileDrop

github_url = solara.util.github_url(__file__)
if sys.platform != "emscripten":
    pycafe_url = solara.util.pycafe_url(path=pathlib.Path(__file__), requirements=["pandas", "plotly"])
else:
    pycafe_url = None
df_sample = pd.read_csv("https://raw.githubusercontent.com/plotly/datasets/master/gapminderDataFiveYear.csv")


class State:
    size_max = solara.reactive(40.0)
    size = solara.reactive(cast(Optional[str], None))
    color = solara.reactive(cast(Optional[str], None))
    x = solara.reactive(cast(Optional[str], None))
    y = solara.reactive(cast(Optional[str], None))
    logx = solara.reactive(False)
    logy = solara.reactive(False)
    df = solara.reactive(cast(Optional[pd.DataFrame], None))

    @staticmethod
    def load_sample():
        State.x.value = "gdpPercap"
        State.y.value = "lifeExp"
        State.size.value = "pop"
        State.color.value = "continent"
        State.logx.value = True
        State.df.value = df_sample

    @staticmethod
    def load_from_file(file):
        df = pd.read_csv(file["file_obj"])
        State.x.value = str(df.columns[0])
        State.y.value = str(df.columns[1])
        State.size.value = str(df.columns[2])
        State.color.value = str(df.columns[3])
        State.df.value = df

    @staticmethod
    def reset():
        State.df.value = None


@solara.component
def Page():
    df = State.df.value
    dark_effective = solara.lab.use_dark_effective()

    # the .scatter will set this cross filter
    filter, _set_filter = solara.use_cross_filter(id(df))

    # only apply the filter if the filter or dataframe changes
    def filter_df():
        if filter is not None and df is not None:
            return df.loc[filter]

    dff = solara.use_memo(filter_df, dependencies=[df, filter])

    with solara.AppBar():
        solara.lab.ThemeToggle()
    with solara.Sidebar():
        with solara.Card("Controls", margin=0, elevation=0):
            with solara.Column():
                with solara.Row():
                    solara.Button("Sample dataset", color="primary", text=True, outlined=True, on_click=State.load_sample, disabled=df is not None)
                    solara.Button("Clear dataset", color="primary", text=True, outlined=True, on_click=State.reset)
                if sys.platform != "emscripten":
                    FileDrop(on_file=State.load_from_file, on_total_progress=lambda *args: None, label="Drag file here")
                else:
                    solara.Info("File upload not supported in this environment")

                if df is not None:
                    solara.SliderFloat(label="Size", value=State.size_max, min=1, max=100)
                    solara.Checkbox(label="Log x", value=State.logx)
                    solara.Checkbox(label="Log y", value=State.logy)
                    columns = list(map(str, df.columns))
                    solara.Select("Column x", values=columns, value=State.x)
                    solara.Select("Column y", values=columns, value=State.y)
                    solara.Select("Size", values=columns, value=State.size)
                    solara.Select("Color", values=columns, value=State.color)
                    if filter is None:
                        solara.Info("If you select points in the scatter plot, you can download the points here.")
                    else:

                        def get_data():
                            return dff.to_csv(index=False)

                        solara.FileDownload(get_data, label=f"Download {len(dff):,} selected points", filename="selected.csv")

    if df is not None:
        with Columns(widths=[2, 4]):
            if State.x.value and State.y.value:
                solara_px.scatter(
                    df,
                    State.x.value,
                    State.y.value,
                    size=State.size.value,
                    color=State.color.value,
                    size_max=State.size_max.value,
                    log_x=State.logx.value,
                    log_y=State.logy.value,
                    template="plotly_dark" if dark_effective else "plotly_white",
                )
            else:
                solara.Warning("Select x and y columns")

    else:
        solara.Info("No data loaded, click on the sample dataset button to load a sample dataset, or upload a file.")

    with solara.Column(style={"max-width": "400px"}):
        solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)
        if sys.platform != "emscripten":
            solara.Button(
                label="Edit this example live on py.cafe",
                icon_name="mdi-coffee-to-go-outline",
                attributes={"href": pycafe_url, "target": "_blank"},
                text=True,
                outlined=True,
            )


@solara.component
def Layout(children):
    route, routes = solara.use_route()
    dark_effective = solara.lab.use_dark_effective()
    return solara.AppLayout(children=children, toolbar_dark=dark_effective, color=None)  # if dark_effective else "primary")



import solara
import solara.lab

github_url = solara.util.github_url(__file__)


@solara.component
def Page():
    with solara.Column():
        solara.Title("I'm in the browser tab and the toolbar bar")
        with solara.Sidebar():
            with solara.Card("I am in the sidebar"):
                with solara.Column():
                    solara.SliderInt(label="Ideal for placing controls")
                    solara.Button(
                        label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True
                    )

        solara.Info("I'm in the main content area, put your main content here")
        with solara.Card("Use solara.Columns to create relatively sized columns"):
            with solara.Columns([1, 2]):
                solara.Success("I'm in the first column")
                solara.Warning("I'm in the second column, I am twice as wide")

        with solara.Card("Use solara.Column to create a full width column"):
            with solara.Column():
                solara.Success("I'm first in this full with column")
                solara.Warning("I'm second in this full with column")
                solara.Error("I'm third in this full with column")

        with solara.Card("Use solara.ColumnsResponsive to response to screen size"):
            with solara.ColumnsResponsive(6, large=4):
                for i in range(6):
                    solara.Info("two per column on small screens, three per column on large screens")


@solara.component
def Layout(children):
    route, routes = solara.use_route()
    return solara.AppLayout(children=children)



import pytest

pytest.importorskip("vaex")
import bqplot
import ipyvuetify as vw
import ipywidgets
import vaex.datasets

import solara
from solara.components.dataframe import (
    DropdownCard,
    FilterCard,
    HeatmapCard,
    HistogramCard,
    ScatterCard,
    SummaryCard,
    TableCard,
)
from solara.hooks.dataframe import provide_cross_filter, use_cross_filter

df = vaex.datasets.titanic()


def test_histogram_card():
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = use_cross_filter(id(df), "test")
        return solara.Text("dummy")

    @solara.component
    def Test():
        provide_cross_filter()
        with solara.VBox() as main:
            HistogramCard(df, column="sex")
            FilterDummy()
        return main

    widget, rc = solara.render(Test(), handle_error=False)
    figure = rc._find(bqplot.Figure).widget
    bars = figure.marks[0]
    assert bars.x.tolist() == ["female", "male"]
    assert bars.y.tolist() == [466, 843]
    assert set_filter is not None
    set_filter(df["survived"] == True)  # noqa
    assert bars.x.tolist() == ["female", "male"]
    assert bars.y.tolist() == [339, 161]
    bars.selected = [0]
    assert df[filter].sex.unique() == ["female"]


def test_dropdown_card():
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = use_cross_filter(id(df), "test")
        return solara.Text("dummy")

    @solara.component
    def Test(column=None):
        provide_cross_filter()
        with solara.VBox() as main:
            DropdownCard(df, column=column)
            FilterDummy()
        return main

    widget, rc = solara.render(Test(column="sex"), handle_error=False)
    select = rc._find(vw.Select)[0].widget
    result: list = select.items
    result.sort(key=lambda item: item["value"])
    assert result == [{"text": "female", "value": "female"}, {"text": "male", "value": "male"}]
    assert select.v_model is None
    select.v_model = {"text": "female", "value": "female"}
    assert df[filter].sex.unique() == ["female"]
    select.v_model = None
    assert filter is None
    assert set(df.sex.unique()) == {"female", "male"}


def testfilter_card():
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = use_cross_filter(id(df), "test")
        return solara.Text("dummy")

    @solara.component
    def Test(column=None):
        provide_cross_filter()
        with solara.VBox() as main:
            FilterCard(df)
            FilterDummy()
        return main

    widget, rc = solara.render(Test(column="sex"), handle_error=False)
    textfield = rc._find(vw.TextField).widget
    assert textfield.v_model == ""
    textfield.v_model = "str_equals(sex, 'female')"
    assert filter is not None
    assert df[filter].sex.unique() == ["female"]
    textfield.v_model = None
    assert filter is None
    assert set(df.sex.unique()) == {"female", "male"}


def test_summary():
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = use_cross_filter(id(df), "test")
        return solara.Text("dummy")

    @solara.component
    def Test():
        provide_cross_filter()
        with solara.VBox() as main:
            SummaryCard(df)
            FilterDummy()
        return main

    widget, rc = solara.render(Test(), handle_error=False)
    html = rc._find(vw.Html)[1].widget
    assert html.children[0] == "1,309"
    assert set_filter is not None
    set_filter(df.sex == "female")
    assert html.children[0] == "466 / 1,309"


def test_table():
    filter = set_filter = None

    @solara.component
    def Test():
        nonlocal filter, set_filter
        provide_cross_filter()
        filter, set_filter = use_cross_filter(id(df), "test")
        return TableCard(df)

    widget, rc = solara.render_fixed(Test(), handle_error=False)
    output = widget.children[-1].children[-1]
    assert isinstance(output, ipywidgets.Output)
    # we can't test the output since no frontend is connected
    # assert output.outputs == ['a']
    assert set_filter is not None
    set_filter(df.sex == "female")
    # assert output.outputs == ['a']


def test_heatmap():
    filter = set_filter = None

    @solara.component
    def Test():
        nonlocal filter, set_filter
        provide_cross_filter()
        filter, set_filter = use_cross_filter(id(df), "test")
        return HeatmapCard(df, x="age", y="fare", debounce=False)

    widget, rc = solara.render_fixed(Test(), handle_error=False)
    figure = widget.children[-1].children[-1]
    assert isinstance(figure, bqplot.Figure)


def test_scatter():
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = use_cross_filter(id(df), "test")
        return solara.Text("dummy")

    @solara.component
    def Test():
        provide_cross_filter()
        with solara.VBox() as main:
            ScatterCard(df, x="age", y="fare")
            FilterDummy()
        return main

    widget, rc = solara.render(Test(), handle_error=False)
    figure = rc._find(bqplot.Figure).widget
    scatter = figure.marks[0]
    scatter.selected = [0]
    assert filter is not None
    assert len(df[filter]) == 1
    scatter.selected = None
    assert filter is None
    scatter.selected = []
    assert filter is None



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Column(align="end", style={"height": "100%", "justify-content": "center"}):
            solara.Markdown(
                """
                    Team 6 built an automatic FAQ creator, including a Discord bot, which crawls a Discord server, and creates a FAQ page
                    based on the most frequently asked questions in the server. The Data is then fed through [Haystack](https://haystack.deepset.ai/),
                    which summarizes and categorizes the questions, and provides answers to be displayed on the FAQ page.

                    Take a look at the [GitHub Repository](https://github.com/silvhua/solarathon-faq-creator) for instructions on how to run the project.
                """
            )
        with solara.v.Html(tag="video", attributes={"controls": "controls", "autoplay": "autoplay"}, style_="width:100%;"):
            solara.v.Html(
                tag="source",
                attributes={"src": "https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_6/preview.webm", "type": "video/webm"},
            )

    solara.Markdown(
        """
            ## Project Team Members
            * [Simone Frisco](https://www.linkedin.com/in/simonefrisco/), Data Scientist
            * [Silvia Hua](https://www.linkedin.com/in/silviahua), Data Scientist
            * [Arunprasadh Senthil](https://www.linkedin.com/in/arun-prasadh-senthil/), Data Engineer
            * [Janet Mardjuki](https://www.linkedin.com/in/jmardjuki/), Software Developer
            * [Roman Gampert](https://www.linkedin.com/in/roman-gampert-5537b9126/), Product
        """
    )



from pathlib import Path
import pytest
import ipyvuetify as vw
import pandas as pd

try:
    import vaex
except ImportError:
    vaex = None

import solara
import solara.lab
from solara.components.cross_filter import Select, magic_value_missing

HERE = Path(__file__).parent

titanic_url = HERE.parent / "titanic.csv"
df_pandas = pd.read_csv(titanic_url)
if vaex is not None:
    df_vaex = vaex.from_pandas(df_pandas)


@pytest.fixture(params=["pandas", "vaex"] if vaex is not None else ["pandas"])
def df_titanic(request):
    named = {"vaex": df_vaex, "pandas": df_pandas} if vaex is not None else {"pandas": df_pandas}
    return named[request.param]


def test_cross_filter_select(df_titanic):
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = solara.use_cross_filter(id(df_titanic), "test")
        return solara.Text("dummy")

    @solara.component
    def Test(column=None, max_unique=100, multiple=False):
        solara.provide_cross_filter()
        with solara.VBox() as main:
            solara.CrossFilterSelect(df_titanic, column=column, max_unique=max_unique, multiple=multiple)
            FilterDummy()
        return main

    widget, rc = solara.render(Test(column="sex"), handle_error=False)
    select = rc._find(Select).widget
    result: list = select.items
    result.sort(key=lambda item: item["value"])
    assert result == [{"value": "female", "text": "female", "count": 466, "count_max": 466}, {"value": "male", "text": "male", "count": 843, "count_max": 843}]
    assert select.value is None
    select.value = {"value": "female"}
    assert filter is not None

    df = df_titanic[filter]
    assert list(df.sex.unique()) == ["female"]
    select.value = None
    assert filter is None
    assert set(df_titanic.sex.unique()) == {"female", "male"}

    # test a column with missing values
    rc.render(Test(column="cabin", max_unique=200))
    result = select.items
    assert len(result) == 187
    result.sort(key=lambda item: item["value"])
    assert result[0] == {
        "text": "A10",
        "value": "A10",
        "count": 1,
        "count_max": 1,
    }
    assert result[-1] == {
        "text": "NA",
        "value": magic_value_missing,
        "count": 1014,
        "count_max": 1014,
    }
    assert select.value is None

    select.value = {"value": magic_value_missing}
    assert filter is not None
    df = df_titanic[filter]
    assert list(df.cabin.unique()) in [[None], []]

    # changing column should clear filter
    rc.render(Test(column="boat"))
    assert filter is None

    # test a column with missing values
    rc.render(Test(column="sex", multiple=True))
    assert select is rc._find(Select).widget
    select.value = [{"value": "female"}]
    assert filter is not None
    df = df_titanic[filter]
    assert len(df) == 466

    select.value = [{"value": "female"}, {"value": "male"}]
    assert filter is not None
    df = df_titanic[filter]
    assert len(df) == 1309


def test_cross_filter_slider(df_titanic):
    filter = set_filter = None

    @solara.component
    def FilterDummy():
        nonlocal filter, set_filter
        filter, set_filter = solara.use_cross_filter(id(df_titanic), "test")
        return solara.Text("dummy")

    @solara.component
    def Test(column=None, max_unique=100, multiple=False):
        solara.provide_cross_filter()
        with solara.VBox() as main:
            solara.CrossFilterSlider(df_titanic, column=column)
            FilterDummy()
        return main

    widget, rc = solara.render(Test(column="pclass"), handle_error=False)
    assert filter is not None
    rc._find(vw.Alert).assert_empty()
    slider = rc._find(vw.Slider).widget
    assert slider.v_model == 1
    assert slider.min == 1
    assert slider.max == 3
    df = df_titanic[filter]
    assert df["pclass"].unique() == [1]

    slider.v_model = 2
    df = df_titanic[filter]
    assert df["pclass"].unique() == [2]

    rc.render(Test(column="age"))
    # will render a different slider
    slider = rc._find(vw.Slider).widget
    assert slider.v_model == 0.1667
    assert slider.min == 0.1667
    assert slider.max == 80.0



import gc
import threading
import time
import weakref
from pathlib import Path
from typing import Optional

import objgraph
import playwright.sync_api
import pytest

import solara
import solara.server.kernel_context

HERE = Path(__file__).parent


set_value = None
context: Optional["solara.server.kernel_context.VirtualKernelContext"] = None


@pytest.fixture
def no_cull_timeout():
    cull_timeout_previous = solara.server.settings.kernel.cull_timeout
    solara.server.settings.kernel.cull_timeout = "0.0001s"
    try:
        yield
    finally:
        solara.server.settings.kernel.cull_timeout = cull_timeout_previous


def _scoped_test_memleak(
    page_session: playwright.sync_api.Page,
    solara_server,
    solara_app,
    extra_include_path,
):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Documentation").first.wait_for()
        assert len(solara.server.kernel_context.contexts) == 1
        context = weakref.ref(list(solara.server.kernel_context.contexts.values())[0])
        # we should not have created a new context
        assert len(solara.server.kernel_context.contexts) == 1
        ctx = context()
        assert ctx is not None
        kernel = weakref.ref(ctx.kernel)
        krn = kernel()
        assert krn is not None
        shell = weakref.ref(krn.shell)
        session = weakref.ref(krn.session)
        last_cull_task = ctx._last_kernel_cull_task
        page_session.goto("about:blank")
        if last_cull_task is not None and not last_cull_task.done():
            event = threading.Event()
            last_cull_task.add_done_callback(lambda _: event.set())
            assert event.wait()
        closed_event = ctx.closed_event
        assert closed_event.wait(10)
        del ctx, krn, last_cull_task, closed_event
        if shell():
            del shell().__dict__
    return context, kernel, shell, session


def test_memleak(
    pytestconfig,
    request,
    browser: playwright.sync_api.Browser,
    page_session: playwright.sync_api.Page,
    solara_server,
    solara_app,
    extra_include_path,
    no_cull_timeout,
):
    # for unknown reasons, del does not work in CI
    context_ref, kernel_ref, shell_ref, session_ref = _scoped_test_memleak(page_session, solara_server, solara_app, extra_include_path)
    # Clears playwright traces, which otherwise hold references to coroutines executed in _scoped_test_memleak
    page_session.context.tracing.stop()

    for i in range(200):
        time.sleep(0.1)
        for gen in [2, 1, 0]:
            gc.collect(gen)
        if context_ref() is None and kernel_ref() is None and shell_ref() is None and session_ref() is None:
            break
    else:
        import sys

        # Otherwise we also see references to NoneType objects
        refs_to_show = [ref() for ref in [context_ref, kernel_ref, shell_ref, session_ref] if ref() is not None]
        name = solara_server.__class__.__name__
        output_path = Path(pytestconfig.getoption("--output")) / f"mem-leak-{name}-python-{sys.version_info}.pdf"
        output_path.parent.mkdir(parents=True, exist_ok=True)
        print("output to", output_path, output_path.resolve())  # noqa
        objgraph.show_backrefs(refs_to_show, filename=str(output_path), max_depth=15)

    assert context_ref() is None
    assert kernel_ref() is None
    assert shell_ref() is None
    assert session_ref() is None



"""
# Card

"""

import solara
import solara.lab
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "Card"

Page = NoPage

__doc__ += apidoc(solara.Card.f)  # type: ignore



from pathlib import Path

import solara
from solara.website.components.markdown import MarkdownWithMetadata

title = "FAQ"
HERE = Path(__file__)


@solara.component
def Page(route_external=None):
    MarkdownWithMetadata(Path(HERE.parent / "content" / "99-faq.md").read_text())



import threading
import time

import solara


def test_display_in_thread(kernel_context):
    rendered_in_thread = threading.Event()
    column = None
    render_count = 0

    @solara.component
    def Page():
        nonlocal column
        nonlocal render_count
        render_count += 1
        force_redraw = solara.use_reactive(0)

        def run():
            time.sleep(0.1)
            force_redraw.value = 1

        solara.use_thread(run, dependencies=[])

        def set_event():
            if force_redraw.value == 1:
                rendered_in_thread.set()

        solara.use_effect(set_event, dependencies=[force_redraw.value])

        s = solara.SliderInt("test")
        with solara.Column() as current_column:
            solara.display("test")
            solara.Button("test")
            solara.display(s)
        if force_redraw.value == 1:
            column = current_column

    _, rc = solara.render(Page(), handle_error=False)
    rendered_in_thread.wait(timeout=2)
    assert column is not None
    # assert render_count == 3
    assert len(column.kwargs["children"]) == 3



import solara


@solara.component
def Sidebar():
    route_current, all_routes = solara.use_route(-1)
    router = solara.use_router()
    if route_current is None:
        return solara.Error("Page not found")

    # Pick out documentation route, so we can use this on any page
    for route in all_routes:
        if route.path == "documentation":
            all_routes = route.children
            break

    with solara.v.List(
        expand=True,
        nav=True,
        style_="height: 100%; max-height: 100vh; display: flex; flex-direction: column; background-color: var(--color-material-background); overflow-y: auto;",
    ) as main:
        with solara.v.ListItemGroup(v_model=router.path):
            # e.g. getting_started, examples, components, api, advanced, faq
            for route in all_routes:
                if len(route.children) == 1 or route.path == "/":
                    with solara.Link("/documentation/" + route.path if route.path != "/" else "/documentation"):
                        with solara.v.ListItem(value="/documentation/" + route.path if route.path != "/" else "/documentation"):
                            if route.path == "/":
                                solara.v.ListItemIcon(children=[solara.v.Icon(children=["mdi-home"])])
                            solara.v.ListItemTitle(style_="padding: 0 20px;", children=[route.label])
                else:
                    path_top_level = "/documentation/" + route.path
                    top_level_expanded = router.path.startswith(path_top_level)
                    with solara.v.ListGroup(
                        v_slots=[
                            {
                                "name": "activator",
                                "children": solara.v.ListItemTitle(
                                    children=[route.label],
                                    style_="padding: 0 20px;",
                                ),
                            }
                        ],
                        value=top_level_expanded,
                        eager=True,  # better for SEO
                    ):
                        for item in route.children:
                            label = item.label
                            if item.path == "/" and route.path in ["examples", "api", "components"]:
                                # the 'homepage' of the subpage are named Overview
                                label = "Overview"
                            path_sub = "/documentation/" + route.path + "/" + item.path
                            sub_should_be_expanded = router.path.startswith(path_sub)
                            if item.children != [] and any([c.label is not None and c.path != "/" for c in item.children]):
                                with solara.v.ListGroup(
                                    v_slots=[
                                        {
                                            "name": "activator",
                                            "children": solara.v.ListItemTitle(
                                                children=[label],
                                            ),
                                        }
                                    ],
                                    sub_group=True,
                                    no_action=True,
                                    eager=True,  # better for SEO
                                    value=sub_should_be_expanded,
                                ):
                                    for subitem in item.children:
                                        # skip the 'homepage' of the examples only
                                        if subitem.path == "/" and route.path not in ["getting_started", "advanced"]:
                                            continue
                                        path = (
                                            "/documentation/" + route.path + "/" + item.path + "/" + (subitem.path if subitem.path != "/" else "")
                                            if item.path != "fullscreen"
                                            else "/apps/" + subitem.path
                                        )
                                        with solara.Link(
                                            path,
                                        ):
                                            with solara.v.ListItem(dense=True, style_="margin-left: 40px; padding: 0 20px;", value=path):
                                                solara.v.ListItemContent(
                                                    children=[subitem.label],
                                                )
                            else:
                                path = "/documentation/" + route.path + ("/" + item.path if item.path != "/" else "")
                                with solara.Link(path):
                                    with solara.v.ListItem(dense=True, style_="padding: 0 20px;", value=path):
                                        solara.v.ListItemContent(
                                            children=[label],
                                        )
            with solara.Link("/contact"):
                with solara.v.ListItem(value="/contact"):
                    solara.v.ListItemIcon(children=[solara.v.Icon(children=["mdi-email"])])
                    solara.v.ListItemTitle(style_="padding: 0 20px;", children=["Contact"])
            with solara.Link("/changelog"):
                with solara.v.ListItem(value="/changelog"):
                    solara.v.ListItemIcon(children=[solara.v.Icon(children=["mdi-history"])])
                    solara.v.ListItemTitle(style_="padding: 0 20px;", children=["Changelog"])
            with solara.Link("/roadmap"):
                with solara.v.ListItem(value="/roadmap"):
                    solara.v.ListItemIcon(children=[solara.v.Icon(children=["mdi-road"])])
                    solara.v.ListItemTitle(style_="padding: 0 20px;", children=["Roadmap"])

    return main



import subprocess
from pathlib import Path

import playwright.sync_api

import solara.server.server

from . import conftest

HERE = Path(__file__).parent


def test_run_widget(page_session: playwright.sync_api.Page):
    port = conftest.TEST_PORT
    conftest.TEST_PORT += 1
    args = ["solara", "run", f"--port={port}", "--no-open", str(HERE / "app_widget.py") + ":button"]
    popen = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
    host = "localhost"
    try:
        solara.server.server.wait_ready(f"http://{host}:{port}", timeout=15)
        page_session.wait_for_timeout(2000)  # on windows, the server is not ready yet?
        page_session.goto(f"http://localhost:{port}")
        page_session.locator("text=Clicked 0 times").click()
        page_session.locator("text=Clicked 1 times").click()
        page_session.locator("text=Clicked 2 times").wait_for()
        popen.kill()
    except Exception as e:
        try:
            popen.kill()
        except:  # noqa
            pass
        outs, errs = popen.communicate(timeout=5)
        if errs:
            print("STDERR:")  # noqa
            print(errs.decode("utf-8"))  # noqa
        if outs:
            print("STDOUT:")  # noqa
            print(outs.decode("utf-8"))  # noqa
        if errs:
            raise ValueError("Expected no errors in solara server output") from e
        raise



import datetime as dt
from typing import List
from unittest.mock import MagicMock

from playwright.sync_api import Page, expect

date = dt.date(2018, 9, 1)
date2 = date + dt.timedelta(days=1)


def test_input_date_single(solara_test, page_session: Page):
    import solara
    from solara.lab import InputDate

    on_value = MagicMock()

    @solara.component
    def Page():
        def update_value(value: dt.date):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(date)
        InputDate(value=value, label="label", on_value=update_value, classes=["test-class"])

    solara.display(Page())

    input = page_session.locator(".test-class input")
    expect(input).to_be_visible()
    expect(input).to_have_value(date.strftime("%Y/%m/%d"))
    input.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).to_be_visible()
    today_button = page_session.get_by_role("button", name=date.strftime("%d").lstrip("0"), exact=True)
    # We click it, but it does not trigger a change, so we don't auto close
    # Do we want to change this behaviour, and still close it?
    today_button.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).to_be_visible()
    tomorrow_button = page_session.get_by_role("button", name=date2.strftime("%d").lstrip("0"), exact=True)
    tomorrow_button.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).not_to_be_visible()

    input.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).to_be_visible()
    page_session.mouse.click(400, 400)
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).not_to_be_visible()

    page_session.wait_for_timeout(350)
    input.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).to_be_visible()
    page_session.keyboard.press("Tab")
    expect(page_session.get_by_role("menu")).not_to_be_visible()


def test_input_date_range(solara_test, page_session: Page):
    import solara
    from solara.lab import InputDateRange

    on_value = MagicMock()

    @solara.component
    def Page():
        def update_value(value: List[dt.date]):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state([date, date2])
        InputDateRange(value=value, label="label", on_value=update_value, classes=["test-class"])

    solara.display(Page())

    input = page_session.locator(".test-class input")
    expect(input).to_be_visible()
    expect(input).to_have_value(f"{date.strftime('%Y/%m/%d')} - {date2.strftime('%Y/%m/%d')}")
    input.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).to_be_visible()
    today_button = page_session.get_by_role("button", name=date.strftime("%d").lstrip("0"), exact=True)
    today_button.click()
    page_session.wait_for_timeout(350)
    expect(page_session.locator(".test-class label")).to_contain_text("label (Please select two dates)")
    expect(page_session.get_by_role("menu")).to_be_visible()
    tomorrow_button = page_session.get_by_role("button", name=date2.strftime("%d").lstrip("0"), exact=True)
    tomorrow_button.click()
    page_session.wait_for_timeout(350)
    expect(page_session.locator(".test-class label")).not_to_contain_text("(Please select two dates)")
    expect(page_session.get_by_role("menu")).not_to_be_visible()
    input.click()
    page_session.mouse.click(400, 400)
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).not_to_be_visible()

    page_session.wait_for_timeout(350)
    input.click()
    page_session.wait_for_timeout(350)
    expect(page_session.get_by_role("menu")).to_be_visible()
    page_session.keyboard.press("Tab")
    expect(page_session.get_by_role("menu")).not_to_be_visible()



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Column(align="end", style={"height": "100%", "justify-content": "center"}):
            solara.Markdown(
                """
                [Wanderlust](https://huggingface.co/spaces/solara-dev/wanderlust) is a reproduction of the travel assistant demo shown at the
                [OpenAI DevDay](https://devday.openai.com/) 2023, built using Solara and the OpenAI Assistants API.

                The sourcecode of the demo is available on [github](https://github.com/widgetti/wanderlust)
                """
            )
        with solara.v.Html(tag="a", style_="width: 100%;", attributes={"href": "https://huggingface.co/spaces/solara-dev/wanderlust", "target": "_blank"}):
            solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/wanderlust/thumbnail.png", width="100%")
    with solara.ColumnsResponsive(12, medium=6):
        with solara.v.Html(tag="a", style_="width: 100%;", attributes={"href": "https://huggingface.co/spaces/solara-dev/wanderlust", "target": "_blank"}):
            solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/wanderlust/assistant.png", width="100%")
        with solara.Column(align="end", style={"height": "100%", "justify-content": "center"}):
            solara.Markdown(
                """Using the new [function calling feature](https://platform.openai.com/docs/guides/function-calling) of OpenAI Assistants, the
                            assistant can annotate, move, and zoom the map to provide additional functionality"""
            )



import solara


@solara.component
def Page():
    with solara.Column(gap="40px", style={"max-width": "90%"}):
        solara.Markdown("""
# Our Team

Solara is primarily developed by [Widgetti](https://widgetti.io/), a software development company specializing in the front-end side of the jupyter ecosystem. Widgetti was founded by Maarten Breddels and Mario Buikhuizen in 2023, and is based in the Netherlands.

## Core Team
        """)
        DevCard(
            "Maarten Breddels",
            "Co-Founder of Widgetti, Creator of Solara",
            "https://github.com/maartenbreddels",
            "Maarten is the creator of Solara, ipyvolume, and a founder of Widgetti. He is a core developer of the [Vaex](https://vaex.io/) project, and has a background in working with large datasets in astronomy.",
            image="https://dxhl76zpt6fap.cloudfront.net/public/avatar/maarten-breddels.jpeg",
            linkedin="https://www.linkedin.com/in/maartenbreddels/",
            twitter="https://twitter.com/maartenbreddels",
            email="maartenbreddels@widgetti.io",
        )
        DevCard(
            "Mario Buikhuizen",
            "Co-Founder of Widgetti",
            "https://github.com/mariobuikhuizen",
            "Mario is a co-founder of Widgetti and has a background in software development and data engineering. He is the creator of [ipyvuetify](https://github.com/widgetti/ipyvuetify) and [ipyvue](https://github.com/widgetti/ipyvue).",
            image="https://dxhl76zpt6fap.cloudfront.net/public/avatar/mario-buikhuizen.jpeg",
            linkedin="https://www.linkedin.com/in/mariobuikhuizen/",
            twitter="https://twitter.com/mariobuikhuizen",
            email="mariobuikhuizen@widgetti.io",
        )
        DevCard(
            "Iisakki Rotko",
            "Medior Software Engineer",
            "https://github.com/iisakkirotko",
            "Iisakki is a medior software engineer at Widgetti, working primarily on Solara. He has a background in physics and web development.",
            image="https://dxhl76zpt6fap.cloudfront.net/public/avatar/iisakki-rotko.jpg",
            linkedin="https://www.linkedin.com/in/iisakkirotko/",
            email="iisakki.rotko@widgetti.io",
        )


@solara.component
def DevCard(
    name,
    role,
    github,
    bio,
    image=None,
    linkedin=None,
    twitter=None,
    website=None,
    email=None,
):
    with solara.Row(gap="40px", style={"align-items": "stretch", "max-height": "300px", "flex": "1 1 auto"}):
        with solara.Div(style={"flex": "0 1 300px", "border-radius": "15px", "overflow": "hidden", "aspect-ratio": "1/1"}):
            solara.v.Html(
                tag="img",
                attributes={"src": image or "https://dxhl76zpt6fap.cloudfront.net/public/logo.svg", "alt": name},
                style_="height: 100%; aspect-ratio: 1; object-fit: cover;",
            )
        with solara.Column():
            solara.HTML(tag="h2", unsafe_innerHTML=name)
            solara.HTML(tag="h3", unsafe_innerHTML=role)
            solara.Markdown(bio)
            with solara.Row(gap="10px"):
                if linkedin:
                    with solara.v.Html(tag="a", attributes={"href": linkedin, "target": "_blank"}):
                        solara.v.Icon(children=["mdi-linkedin"], color="var(--color-grey-light)")
                if twitter:
                    with solara.v.Html(tag="a", attributes={"href": twitter, "target": "_blank"}):
                        solara.v.Icon(children=["mdi-twitter"], color="var(--color-grey-light)")
                if website:
                    with solara.v.Html(tag="a", attributes={"href": website, "target": "_blank"}):
                        solara.v.Icon(children=["mdi-web"], color="var(--color-grey-light)")
                if email:
                    with solara.v.Html(tag="a", attributes={"href": f"mailto:{email}"}):
                        solara.v.Icon(children=["mdi-email"], color="var(--color-grey-light)")
                if github:
                    with solara.v.Html(tag="a", attributes={"href": github, "target": "_blank"}):
                        solara.v.Icon(children=["mdi-github"], color="var(--color-grey-light)")



import ipyvuetify as vy
import solara


class CodeHighlightCssWidget(vy.VuetifyTemplate):
    template_file = (__file__, "code_highlight_css.vue")


@solara.component
def CodeHighlightCss():
    return CodeHighlightCssWidget.element()



import logging
import mimetypes

import tornado.web
from jupyter_server.base.handlers import JupyterHandler

from solara.server import settings
from solara.server.cdn_helper import get_data

logger = logging.getLogger("Solara.cdn")


class CdnHandler(JupyterHandler):
    def initialize(self, cache_directory=settings.assets.proxy_cache_dir):
        self.cache_directory = cache_directory
        logging.info("Using %r as cache directory", self.cache_directory)

    async def get(self, path=None):
        try:
            content = get_data(self.cache_directory, path)
        except Exception as e:
            logger.warning(e)
            raise tornado.web.HTTPError(500)

        mime = mimetypes.guess_type(path)
        if mime[0] is not None:
            self.set_header("Content-Type", mime[0])
        self.write(content)



import json
import os
import requests

from typing import Any, Dict, Optional
import solara


postmark_api_key = None
contact_email_address = None

try:
    postmark_api_key = os.environ["POSTMARK_API_KEY"]
    contact_email_address = os.environ["SOLARA_CONTACT_EMAIL_ADDRESS"]
except Exception:
    pass


@solara.component
def Contact(
    style: Dict[str, Any] = {},
    title="Contact Us",
    subtitle="We'd love to hear from you!",
    submit_label="Submit",
    email_subject="Contact Form Submission",
):
    first_name = solara.use_reactive("")
    last_name = solara.use_reactive("")
    email = solara.use_reactive("")
    company = solara.use_reactive("")
    message = solara.use_reactive("")
    error: solara.Reactive[Optional[str]] = solara.use_reactive(None)
    success = solara.use_reactive(False)

    def send(*_ignore):
        if postmark_api_key is None or contact_email_address is None:
            error.set("Email service not properly configured. Please contact the site administrator at solara@widgetti.io.")
        elif not first_name.value or not last_name.value or not email.value or not message.value:
            error.set("Please fill out all required fields.")
        else:
            # Create the email content
            msg = {}
            msg["From"] = contact_email_address
            msg["To"] = contact_email_address
            msg["Subject"] = email_subject
            msg["ReplyTo"] = email.value

            # Email body
            msg["HtmlBody"] = f"""
            <b>First Name</b>: {first_name.value}<br />
            <b>Last Name</b>: {last_name.value}<br />
            <b>Email</b>: {email.value}<br />
            <b>Company</b>: {company.value}<br />
            <b>Message</b>: {message.value}<br />
            """

            # Send emails
            try:
                requests.post(
                    "https://api.postmarkapp.com/email",
                    headers={
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                        "X-Postmark-Server-Token": postmark_api_key,
                    },
                    data=json.dumps(msg),
                )
                requests.post(
                    "https://api.postmarkapp.com/email",
                    headers={
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                        "X-Postmark-Server-Token": postmark_api_key,
                    },
                    data=json.dumps(
                        {
                            "From": contact_email_address,
                            "To": email.value,
                            "Subject": "Thank you for contacting Solara",
                            "HtmlBody": f"""
                            <p>Hi {first_name.value},</p>
                            <p>Thank you for contacting us! We will get back to you as soon as possible.</p>
                            <p>Best regards,<br />The Solara Team</p>
                            """,
                        }
                    ),
                )
            except Exception as e:
                error.set(f"Error sending email: {e}")
            else:
                success.set(True)
                error.set(None)
                first_name.set("")
                last_name.set("")
                email.set("")
                company.set("")
                message.set("")

    with solara.Card(title=title, style={"width": "100%", "max-width": "1024px", **style}):
        solara.Markdown(subtitle)
        solara.Text("* Required fields")
        with solara.Row():
            solara.InputText(label="First Name *", value=first_name)
            solara.InputText(label="Last Name *", value=last_name)
        with solara.Row():
            solara.InputText(label="Email *", value=email)
            solara.InputText(label="Company", value=company)
        solara.v.Textarea(placeholder="Message *", v_model=message.value, on_v_model=message.set)
        with solara.CardActions():
            solara.Button(label=submit_label, color="primary", on_click=send)
            solara.Button(
                label="Clear",
                color="secondary",
                text=True,
                on_click=lambda: [first_name.set(""), last_name.set(""), email.set(""), company.set(""), message.set("")],
            )

    solara.Style(
        """
        .v-snack__wrapper {
            box-shadow: none;
        }
        """
    )

    with solara.v.Snackbar(
        v_model=error.value is not None,
        timeout=5000,
        on_v_model=lambda *_: error.set(None),
        left=True,
        color="error",
    ):
        solara.Markdown(error.value or "", style={"--dark-color-text": "white", "--color-text": "white"})
        solara.Button(icon=True, icon_name="mdi-close", color="white", on_click=lambda: error.set(None))

    with solara.v.Snackbar(
        v_model=success.value,
        timeout=5000,
        on_v_model=lambda *_: success.set(False),
        left=True,
        color="success",
    ):
        solara.Markdown("Your message has been sent!", style={"--dark-color-text": "white", "--color-text": "white"})
        solara.Button(icon=True, icon_name="mdi-close", color="white", on_click=lambda: success.set(False))



import operator
import os
from functools import reduce
from typing import Any, Callable, Dict, List, NoReturn

import ipyvuetify as v

try:
    import numpy as np
except ModuleNotFoundError:
    np = None  # type: ignore
import traitlets

import solara
from solara.alias import rv
from solara.lab.hooks.dataframe import use_df_column_names

cardheight = "100%"


class PivotTableWidget(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "pivot_table.vue"))
    d = traitlets.Dict(default_value={"no": "data"}).tag(sync=True)
    selected = traitlets.Dict(default_value={}).tag(sync=True)
    style_ = traitlets.Unicode("").tag(sync=True)


def assert_never(value) -> NoReturn:
    # This also works at runtime as well
    assert False, f"This code should never be reached, got: {value}"


def translate_agg_to_vaex(aggregation: solara.Aggregation, filter=None):
    if aggregation["type"] == "count":
        import vaex

        return vaex.agg.count(selection=filter) if filter else vaex.agg.count()
    assert_never(aggregation)


def df_aggregate_vaex(df, columns: List[str], aggregations: Dict[str, solara.Aggregation], filter=None):
    aggs = {key: translate_agg_to_vaex(agg, filter) for key, agg in aggregations.items()}
    return df.groupby(columns, sort=True).agg(aggs)


def df_aggregate_pivot_vaex(df, x: List[str], y: List[str], aggregation: solara.Aggregation, filter=None) -> solara.PivotTableData:
    agg_name = aggregation["type"]

    columns = [*x, *y]
    agg_column = "agg_count__"
    row_index = "__row_index__"

    aggregations = {agg_column: aggregation}
    if x and y:
        dfg = df_aggregate_vaex(df, columns, aggregations, filter=filter)
    else:
        dfg = None
    if x:
        dfgx = df_aggregate_vaex(df, x, aggregations, filter=filter)
        dfgx[row_index] = np.arange(len(dfgx), dtype="int64")
    else:
        dfgx = None
    if y:
        dfgy = df_aggregate_vaex(df, y, aggregations, filter=filter)
        dfgy[row_index] = np.arange(len(dfgy), dtype="int64")
    else:
        dfgy = None
    dfg_total = df._agg(translate_agg_to_vaex(aggregations[agg_column], filter=filter))

    def formatter(value):
        if isinstance(value, float):
            return f"{value:,.2f}"
        else:
            return f"{value:,d}"

    if x and y:
        values: List[List[solara.JsonType]] = [[None] * len(dfgy) for _ in range(len(dfgx))]
    else:
        values = [[]]
    if x:
        values_x = list(map(formatter, dfgx[agg_column].tolist()))
        headers_x = [dfgx[k].tolist() for k in x]
        counts_x = len(dfgx)
    else:
        values_x = []
        headers_x = []
        counts_x = 0
    if y:
        values_y = list(map(formatter, dfgy[agg_column].tolist()))
        headers_y = [dfgy[k].tolist() for k in y]
        counts_y = len(dfgy)
    else:
        values_y = []
        headers_y = []
        counts_y = 0
    total = formatter(dfg_total)

    if x and y:
        for row in dfg.to_records():
            dfs = {"x": dfgx, "y": dfgy}
            for column in columns:
                axis_name = "x" if column in x else "y"
                df_axis = dfs[axis_name]
                if row[column] is None:
                    df_axis = df_axis[df[column].ismissing()]
                else:
                    df_axis = df_axis[df_axis[column] == row[column]]
                dfs[axis_name] = df_axis
            assert len(dfs["x"])
            assert len(dfs["y"])
            xi = dfs["x"][row_index].tolist()[0]
            yi = dfs["y"][row_index].tolist()[0]
            value = row[agg_column]
            values[xi][yi] = formatter(value)

    data: solara.PivotTableData = {
        "x": x,
        "y": y,
        "agg": agg_name,
        "values": values,
        "values_x": values_x,
        "values_y": values_y,
        "headers_x": headers_x,
        "headers_y": headers_y,
        "counts_x": counts_x,
        "counts_y": counts_y,
        "total": total,
    }
    return data


def use_df_pivot_data(df, x: List[str], y: List[str], aggregation: solara.Aggregation, filter=None) -> solara.Result[solara.PivotTableData]:
    return solara.use_thread(lambda: df_aggregate_pivot_vaex(df, x, y, aggregation, filter=filter), dependencies=[*x, *y, aggregation, filter])


@solara.component
def PivotTableView(data: solara.PivotTableData, selected: Dict[str, Any] = {}, on_selected: Callable[[Dict[str, Any]], None] = None, style=""):
    return PivotTableWidget.element(d=data, selected=selected, on_selected=on_selected, style_=style)


@solara.component
def PivotTable(
    df,
    x: List[str] = [],
    y: List[str] = [],
    aggregation: solara.Aggregation = solara.AggregationCount(type="count"),
    selected: Dict[str, Any] = {},
    on_selected: Callable[[Dict[str, Any]], None] = None,
):
    x = x.copy()
    y = y.copy()
    filter, set_filter = solara.use_cross_filter(id(df), "pivottable")
    dff = df
    data_result = use_df_pivot_data(dff, x, y, aggregation, filter=filter)
    previous = solara.use_previous(data_result.value, condition=data_result.state == solara.ResultState.FINISHED)
    selected, set_selected = solara.use_state_or_update(selected.copy())

    def set_filter_from_pivot_selection(selection):
        set_selected(selection)
        if on_selected:
            on_selected(selection)
        data = data_result.value
        if data is None:
            return
        # assert data is not None
        filters = []
        if "x" in selection:
            sel = selection["x"]
            for level in range(sel[0] + 1):
                value = data["headers_x"][level][sel[1]]
                column = data["x"][level]
                if value is None:
                    filters.append(df[column].ismissing())
                else:
                    filters.append(df[column] == value)

        if "y" in selection:
            sel = selection["y"]
            for level in range(sel[0] + 1):
                column = data["y"][level]
                value = data["headers_y"][level][sel[1]]
                if value is None:
                    filters.append(df[column].ismissing())
                else:
                    filters.append(df[column] == value)
        if filters:
            filter = reduce(operator.and_, filters[1:], filters[0])
        else:
            filter = None
        set_filter(filter if filter is not None else None)

    # Bug in use_thread on state result we remember that state is finished
    if data_result.state == solara.ResultState.FINISHED and data_result.value is not None:
        with solara.VBox() as main:
            rv.ProgressLinear(indeterminate=False, style_="visibility: hidden;")
            PivotTableView(data=data_result.value, selected=selected, on_selected=set_filter_from_pivot_selection)
    elif previous is not None:
        with solara.VBox() as main:
            rv.ProgressLinear(indeterminate=True)
            PivotTableView(data=previous, selected=selected, on_selected=set_filter_from_pivot_selection, style="opacity: 0.3; pointer-events: none;")
    elif data_result.state == solara.ResultState.ERROR:
        return solara.Error(f"Oops: {data_result.error}")
    else:
        return solara.Info(f"Status: {data_result.state}")
    return main


@solara.component
def PivotTableCard(
    df,
    x=[],
    y=[],
    selected: Dict[str, Any] = {},
    on_selected: Callable[[Dict[str, Any]], None] = None,
):
    items = use_df_column_names(df)
    with rv.Card(elevation=2, style_="position: relative", height=cardheight) as main:
        with rv.CardTitle(children=["Pivot table"]):
            pass
        with rv.CardText():
            with rv.Btn(v_on="x.on", icon=True, absolute=True, style_="right: 10px; top: 10px") as btn:
                rv.Icon(children=["mdi-settings"])
            with rv.Dialog(v_slots=[{"name": "activator", "variable": "x", "children": btn}]):
                with rv.Sheet():
                    with rv.Container(pa_4=True, ma_0=True):
                        with rv.Row():
                            with rv.Col():
                                with rv.Card(elevation=2):
                                    with rv.CardTitle(children=["Rows"]):
                                        pass
                                    with rv.CardText():
                                        for i in range(10):
                                            col = solara.ui_dropdown(value=x[i] if i < len(x) else None, label=f"Row {i}", options=items)
                                            if col is None:
                                                break
                                            else:
                                                if i < len(x):
                                                    x[i] = col
                                                else:
                                                    x.append(col)
                            with rv.Col():
                                with rv.Card(elevation=2):
                                    with rv.CardTitle(children=["Columns"]):
                                        pass
                                    with rv.CardText():
                                        for i in range(10):
                                            col = solara.ui_dropdown(value=y[i] if i < len(y) else None, label=f"Column {i}", options=items)
                                            if col is None:
                                                break
                                            else:
                                                if i < len(y):
                                                    y[i] = col
                                                else:
                                                    y.append(col)

            PivotTable(df, x, y, selected=selected, on_selected=on_selected)

    return main



import datetime as dt
from typing import Callable, Dict, List, Optional, Union

import solara
import solara.lab
from solara.lab.components.input_date import use_close_menu
from solara.components.input import _use_input_type


@solara.component
def InputTime(
    value: Union[solara.Reactive[Optional[dt.time]], Optional[dt.time]],
    on_value: Optional[Callable[[Optional[dt.time]], None]] = None,
    label: str = "Pick a time",
    children: List[solara.Element] = [],
    open_value: Union[solara.Reactive[bool], bool] = False,
    on_open_value: Optional[Callable[[bool], None]] = None,
    optional: bool = False,
    twelve_hour_clock: bool = False,
    use_seconds: bool = False,
    allowed_minutes: Optional[List[int]] = None,
    style: Optional[Union[str, Dict[str, str]]] = None,
    classes: Optional[List[str]] = None,
):
    """
    Show a textfield, which when clicked, opens a timepicker. The input time should be of type `datetime.time`.

    ## Basic Example

    ```solara {pycafe-link}
    import solara
    import solara.lab
    import datetime as dt


    @solara.component
    def Page():
        time = solara.use_reactive(dt.time(12, 0))

        solara.lab.InputTime(time, allowed_minutes=[0, 15, 30, 45])
        solara.Text(str(time.value))
    ```

    ## Arguments

    * value: Reactive variable of type `datetime.time`, or `None`. This time is selected the first time the component is rendered.
    * on_value: a callback function for when value changes. The callback function receives the new value as an argument.
    * label: Text used to label the text field that triggers the timepicker.
    * children: List of Elements to be rendered under the timepicker. If empty, a close button is rendered.
    * open_value: Controls and communicates the state of the timepicker. If True, the timepicker is open. If False, the timepicker is closed.
    Intended to be used in conjunction with a custom set of controls to close the timepicker.
    * on_open_value: a callback function for when open_value changes. Also receives the new value as an argument.
    * optional: Determines whether to show an error when value is `None`. If `True`, no error is shown.
    * twelve_hour_clock: If `True`, the timepicker will display in 12-hour format. If `False`, the timepicker will display in 24-hour format.
    * use_seconds: If `True`, the timepicker will allow input of seconds.
    * allowed_minutes: List of allowed minutes for the timepicker. Restricts the input to specific minute intervals.
    * style: CSS style to apply to the text field. Either a string or a dictionary of CSS properties (i.e. `{"property": "value"}`).
    * classes: List of CSS classes to apply to the text field.
    """
    time_format_internal = f"%H:%M{':%S' if use_seconds else ''}"
    time_format_display = f"%H:%M{':%S' if use_seconds else ''}"
    if twelve_hour_clock:
        time_format_display = f"%I:%M{':%S' if use_seconds else ''} %p"
    value_reactive = solara.use_reactive(value, on_value)  # type: ignore
    del value, on_value
    timepicker_is_open = solara.use_reactive(open_value, on_open_value)  # type: ignore
    del open_value, on_open_value

    def set_time_typed_cast(value: Optional[str]):
        if value:
            try:
                time_value = dt.datetime.strptime(value, time_format_display).time()
                return time_value
            except ValueError:
                raise ValueError(f"Time {value} does not match format {time_format_display.replace('%', '')}")
        elif optional:
            return None
        else:
            raise ValueError("Time cannot be empty")

    def time_to_str(time: Optional[dt.time]) -> str:
        if time is not None:
            return time.strftime(time_format_display)
        return ""

    def set_time_cast(new_value: Optional[str]):
        if new_value:
            time_value = dt.datetime.strptime(new_value, time_format_internal).time()
            value_reactive.value = time_value

    def standard_strfy(time: Optional[dt.time]):
        if time is None:
            return None
        else:
            return time.strftime(time_format_internal)

    time_standard_str = standard_strfy(value_reactive.value)

    style_flat = solara.util._flatten_style(style)

    internal_value, error_message, set_value_cast = _use_input_type(value_reactive, set_time_typed_cast, time_to_str)

    if error_message:
        label += f" ({error_message})"
    input = solara.v.TextField(
        label=label,
        v_model=internal_value,
        on_v_model=set_value_cast,
        append_icon="mdi-clock",
        error=bool(error_message),
        style_="min-width: 290px;" + style_flat,
        class_=", ".join(classes) if classes else "",
    )

    use_close_menu(input, timepicker_is_open)

    with solara.lab.Menu(
        activator=input,
        close_on_content_click=False,
        open_value=timepicker_is_open,
        use_activator_width=False,
    ):
        with solara.v.TimePicker(
            ampm_in_title=twelve_hour_clock,
            v_model=time_standard_str,
            on_v_model=set_time_cast,
            format="24hr" if not twelve_hour_clock else "ampm",
            allowed_minutes=allowed_minutes,
            use_seconds=use_seconds,
            style_="width: 100%;",
        ):
            if len(children) > 0:
                solara.display(*children)



import sys
from types import ModuleType
from typing import Any, Dict, Optional

import ipyvuetify as v
import pytest

redis: Optional[ModuleType]
try:
    import redis
except ImportError:
    redis = None

import solara
import solara.cache

from .common import click

some_global = 1


def test_lru():
    c = solara.cache.Memory(max_items=2)
    c["a"] = 1
    c["b"] = 2
    c["c"] = 3
    assert "a" not in c
    assert "b" in c
    assert "c" in c
    c["b"] = 4
    assert "b" in c
    c["d"] = 5
    assert "c" not in c
    assert "b" in c
    assert "d" in c


def test_memoized_basic():
    c = solara.cache.Memory(max_items=2)

    @solara.cache.memoize(storage=c)
    def f(x: int) -> int:
        return x**2

    assert f(1) == 1
    assert f(2) == 4
    assert list(c.values()) == [1, 4]
    assert f(3) == 9
    assert list(c.values()) == [4, 9]


def test_memoized_nonlocals():
    c = solara.cache.Memory(max_items=2)

    some_nonlocal = 1

    @solara.cache.memoize(storage=c, allow_nonlocals=True)
    def f1(x: int) -> int:
        return x**2 + some_nonlocal

    assert f1(2) == 5
    # although some_nonlocal is changed, the cache is not invalidated
    # which means we get the same answer, while some_nonlocal has changed
    some_nonlocal = 10
    assert f1(2) == 5

    with pytest.raises(ValueError):

        @solara.cache.memoize(storage=c)
        def f2(x: int) -> int:
            return x**2 + some_nonlocal


def test_memoized_changing_globals():
    global some_global
    c = solara.cache.Memory(max_items=2)
    with pytest.raises(ValueError):
        # by changing a global, the memoize decorator should fail
        for value in [1, 10]:
            some_global = value

            @solara.cache.memoize(storage=c)
            def g(x: int) -> int:
                return x**2 + some_global

    @solara.cache.memoize(storage=c)
    def h(x: int) -> int:
        return x**2 + some_global

    with pytest.raises(ValueError):
        h(1)
        # by changing a global, the memoized function should fail
        some_global = 20
        h(2)


def test_memoized_naked():
    # no arguments to the decorator
    @solara.cache.memoize
    def f(x: int) -> int:
        return x**2

    f.storage.clear()
    assert f(1) == 1
    assert f(2) == 4
    assert list(f.storage.values()) == [1, 4]
    assert f(3) == 9
    assert list(f.storage.values()) == [1, 4, 9]


def test_memoized_custom_key():
    c = solara.cache.Memory(max_items=2)

    def mykey(x: int) -> str:
        return f"{-x}"

    @solara.cache.memoize(storage=c, key=mykey)
    def f(x: int) -> str:
        return f"{x**2}"

    assert f(1) == "1"
    assert f(2) == "4"
    assert [k[-1] for k in list(c.keys())] == ["-1", "-2"]
    assert list(c.values()) == ["1", "4"]
    assert f(3) == "9"
    assert list(c.values()) == ["4", "9"]
    assert [k[-1] for k in list(c.keys())] == ["-2", "-3"]


def test_memoized_similar_function_name():
    # if we share the cache between two functions with the same name, but different
    # implementations, we should get the correct results
    # this means the there should be something unique in the key related to the function
    @solara.cache.memoize
    def f(x: int) -> int:  # type: ignore
        return x**2

    f1 = f
    del f

    @solara.cache.memoize  # type: ignore
    def f(x: int) -> int:  # type: ignore
        return -(x**2)

    f2 = f

    assert f1(2) == 4
    assert f2(2) == -4


def test_cache_disk(tmpdir):
    path = tmpdir.join("cache")
    c = solara.cache.create("disk", path=str(path), clear=True)
    c["a"] = 1
    assert c["a"] == 1
    assert len(c) == 1
    del c["a"]
    assert "a" not in c
    assert len(c) == 0

    @solara.cache.memoize(storage=c)
    def f(*args, **kwargs):
        return args, kwargs

    # non trivial arguments
    assert f(1, 2, 3, a=1, b=2) == ((1, 2, 3), {"a": 1, "b": 2})
    assert f(1, 2, 3, a=1, b=2, c=3) == ((1, 2, 3), {"a": 1, "b": 2, "c": 3})
    assert f(1, 2, 3, a=1, b=2) == ((1, 2, 3), {"a": 1, "b": 2})
    assert len(c) == 2


@pytest.mark.skipif(condition=redis is None or sys.platform.startswith("win"), reason="redis not installed")
def test_cache_redis(tmpdir):
    c = solara.cache.create("redis", clear=True, prefix=b"solara-test:cache:")
    c["a"] = 1
    assert c["a"] == 1
    assert len(c) == 1
    del c["a"]
    assert "a" not in c
    assert len(c) == 0

    @solara.cache.memoize(storage=c)
    def f(*args, **kwargs):
        return args, kwargs

    # non trivial arguments
    assert f(1, 2, 3, a=1, b=2) == ((1, 2, 3), {"a": 1, "b": 2})
    assert f(1, 2, 3, a=1, b=2, c=3) == ((1, 2, 3), {"a": 1, "b": 2, "c": 3})
    assert f(1, 2, 3, a=1, b=2) == ((1, 2, 3), {"a": 1, "b": 2})
    assert len(c) == 2


def test_cache_memory_size():
    c = solara.cache.create("memory-size", max_size="100b")
    b40 = b"0123456789" * 2  # when pickled ~40 bytes (on py36 less)
    b80 = b40 * 3  # when pickled ~80 bytes
    c["a"] = b40
    c["b"] = b40
    assert list(c.keys()) == [b"a", b"b"]
    c["c"] = b80
    assert list(c.keys()) == [b"c"]

    # test non trivial keys
    c[(1, 2, {"a"})] = b40
    c[(1, 2, {"b"})] = b40
    assert b"c" not in c
    assert len(c) == 2


def test_multi_level_cache():
    l1: Dict[str, Any] = {}
    l2: Dict[str, Any] = {}
    cache = solara.cache.create("multi-level", l1, l2)
    with pytest.raises(KeyError):
        cache["key1"]
    assert l1 == {}
    assert l2 == {}
    # setting should fill all caches
    cache["key1"] = 1
    assert l1 == {"key1": 1}
    assert l2 == {"key1": 1}
    assert cache["key1"] == 1
    del l1["key1"]
    assert l1 == {}
    assert l2 == {"key1": 1}
    # reading should fill l1 as well
    assert cache["key1"] == 1
    assert l1 == {"key1": 1}
    assert l2 == {"key1": 1}

    cache2 = solara.cache.create("memory,disk")
    assert type(cache) is type(cache2)


def test_memoize_hook():
    result_values = []

    @solara.component
    def Test():
        count, set_count = solara.use_state(0)

        @solara.cache.memoize
        def f(x: int) -> int:
            return x**2

        f.intrusive_cancel = False

        result = f.use_thread(10)
        result_values.append(result)
        if result.state == solara.ResultState.FINISHED:
            return solara.Button(str(count), on_click=lambda: set_count(count + 1))
        else:
            return solara.Text("running")

    box, rc = solara.render(Test(), handle_error=False)
    rc.find(v.Btn, children=["0"]).wait_for(timeout=10)
    result_values.clear()
    click(rc.find(v.Btn).widget)
    rc.find(v.Btn, children=["1"]).wait_for(timeout=10)
    assert len(result_values) == 1
    # we should directly get the result from the cache, so we don't go into running state
    assert result_values[0].state == solara.ResultState.FINISHED
    assert result_values[0].value == 100


def test_memoize_hook_no_None_after_hit():
    has_been_none = False

    selected = solara.Reactive("1")

    @solara.memoize
    def something(i):
        return i

    something.intrusive_cancel = False

    @solara.component
    def Test():
        result = something.use_thread(selected.value)

        if result.state == solara.ResultState.FINISHED:
            if result.value is None:
                # this should not happen
                nonlocal has_been_none
                has_been_none = True

            solara.Text(str(result.value))

    box, rc = solara.render(Test(), handle_error=False)
    rc.find(v.Html, children=["1"]).wait_for(timeout=2)

    assert not has_been_none
    selected.set("2")
    rc.find(v.Html, children=["2"]).wait_for(timeout=2)
    assert not has_been_none

    selected.set("1")
    rc.find(v.Html, children=["1"]).wait_for(timeout=2)
    assert not has_been_none

    selected.set("3")
    rc.find(v.Html, children=["3"]).wait_for(timeout=2)
    assert not has_been_none



import logging
import os
from typing import Set

import playwright.sync_api
import pytest

import solara.server.app
import solara.server.server
import solara.server.settings
from solara.server import reload
from solara.server.flask import ServerFlask
from solara.server.starlette import ServerStarlette

reload.reloader.start()
logger = logging.getLogger("solara-test.integration")

worker = os.environ.get("PYTEST_XDIST_WORKER", "gw0")
TEST_PORT = int(os.environ.get("PORT", "18765")) + int(worker[2:])  # up to 18770 is a valid callback for auth0
SERVER = os.environ.get("SOLARA_SERVER")
if SERVER:
    SERVERS = [SERVER]
else:
    SERVERS = ["flask", "starlette"]


urls: Set[str] = set()

timeout = 18  # in seconds, slightly below the  --timeout=20 argument in integration.yml
# allow symlinks on solara+starlette
solara.server.settings.main.mode = "development"


@pytest.fixture(scope="session")
def url_checks():
    yield
    non_localhost = [url for url in urls if not url.startswith("http://localhost")]
    allow_list = [
        "https://user-images.githubusercontent.com",  # logo in readme
        "https://cdnjs.cloudflare.com/ajax/libs/emojione/2.2.7",  # emojione from markdown
        "https://images.unsplash.com",  # portal
        "https://miro.medium.com",  # portal
        "https://dabuttonfactory.com/",  # in markdown, probably temporary
    ]
    non_allow_urls = [url for url in non_localhost if not any(url.startswith(allow) for allow in allow_list)]
    if non_allow_urls:
        msg = "The following URLs were not allowed (non local host, and not in allow list):\n"
        msg += "\n".join(non_allow_urls)
        raise AssertionError(msg)


# see https://github.com/microsoft/playwright-pytest/issues/23
@pytest.fixture
def context(context: playwright.sync_api.BrowserContext, url_checks):
    context.set_default_timeout(timeout * 1000)

    def handle(route, request: playwright.sync_api.Request):
        urls.add(request.url)
        route.continue_()

    # context.route("**/*", handle)
    yield context


@pytest.fixture
def page(page: playwright.sync_api.Page):
    def log(msg):
        print("PAGE LOG:", msg.text)  # noqa
        logger.debug("PAGE LOG: %s", msg.text)

    page.on("console", log)
    return page


def _serve_flask_in_process(port, host):
    from solara.server.flask import app

    app.run(debug=False, port=port, host=host)


server_classes = {
    "flask": ServerFlask,
    "starlette": ServerStarlette,
}

# override the fixure, and also test with flask


@pytest.fixture(params=SERVERS, scope="session")
def solara_server(request):
    server_class = server_classes[request.param]
    global TEST_PORT
    webserver = server_class(TEST_PORT)
    TEST_PORT += 1

    try:
        webserver.serve_threaded()
        webserver.wait_until_serving()
        yield webserver
    finally:
        webserver.stop_serving()


@pytest.fixture()  # type: ignore # noqa
def page(page):  # noqa
    # on CI, it seems that the above context.set_default_timeout(timeout * 1000) does not apply to page
    # so we set it here again. Maybe in other situations the page is created early.. ?
    page.set_default_timeout(timeout * 1000)
    yield page



import threading
import sys
import pytest

from pathlib import Path
from typing import Optional

import playwright.sync_api
from solara_enterprise import ssg

import solara
from solara.server import settings
from solara.server.kernel_context import VirtualKernelContext, get_current_context

HERE = Path(__file__).parent


text_ssg = "# SSG Test"
text_live = "# Live render"
context: Optional[VirtualKernelContext] = None


def set_value(x: str):
    return None


@solara.component
def SSG():
    global set_value
    global context
    value, set_value = solara.use_state(text_ssg)  # type: ignore
    context = get_current_context()
    with solara.HBox() as main:
        solara.Markdown(value)
        solara.Meta(name="description", property="og:description", content="My page description")
    return main


@pytest.mark.skipif(sys.platform == "win32", reason="Skip on Windows, fails with playwright saying navigation was interrupted")
def test_ssg(page_session: playwright.sync_api.Page, solara_server, solara_app, tmpdir):
    global text
    global context

    settings.ssg.build_path = Path(tmpdir) / "build"
    # would be nice if we can get the headed/headless setting from pytest somehow
    # for now, disable the comment below to run in headed mode
    # settings.ssg.headed = True

    def run():
        # run in different thread, since playwright wants
        # its own event loop
        ssg.ssg_crawl(solara_server.base_url)

    with solara_app("tests.integration.ssg_test:SSG"):
        t = threading.Thread(target=run)
        t.start()
        t.join()
        path = settings.ssg.build_path / "index.html"
        assert path.exists()
        html = path.read_text()
        assert ">SSG Test</h1>" in html, "SSG did not render correctly"
        assert "og:description" in html, "SSG did not render meta correctly"
        assert "My page description" in html, "SSG did not render meta correctly"

        page_session.goto(solara_server.base_url)
        page_session.locator('h1:has-text("SSG Test")').wait_for()
        page_session.locator("#pre-rendered-html-present").wait_for(state="detached")
        assert context is not None
        # we need to use set_value with the right context, set_value is not aware of solara's context
        with context:
            set_value(text_live)
        page_session.locator("text=Live render").wait_for()
        context = None



import asyncio
import hashlib
import logging
from pathlib import Path
from typing import Optional, Union, cast

import ipyvuetify as v

try:
    import watchfiles
except ModuleNotFoundError:
    watchfiles = None  # type: ignore

import solara
import solara.server.settings

logger = logging.getLogger("solara.Style")


@solara.component
def Style(value: Union[str, Path] = ""):
    """Add a custom piece of CSS.

    Note that this is considered an advanced feature, and should be used with caution.

    ## Example

    ```python
    from pathlib import Path
    import solara


    @solara.component
    def Page():
        solara.Style(Path("style.css"))
        solara.Button("Click me", classes=["company-style-button"])
    ```

    ## Arguments

    - `value`: The CSS string of CSS file to insert into the page. In development mode (the default)
        when a Path argument is passed, the file will be watched for changes, and the CSS will be reloaded when
        the file changes. In production mode, the CSS will be inserted once, and not reloaded.
    """
    css_content_reloaded, set_css_content_reloaded = solara.use_state(cast(Optional[str], None))

    def hot_reload():
        in_production = solara.server.settings.main.mode == "production"
        if not in_production and isinstance(value, Path):

            async def watch():
                try:
                    async for _ in watchfiles.awatch(value):
                        print(value, "changed, reloading css")  # noqa
                        set_css_content_reloaded(cast(Path, value).read_text(encoding="utf-8"))
                except RuntimeError:
                    pass  # swallow the RuntimeError: Already borrowed errors from watchfiles
                except Exception:
                    logger.exception("Error watching file")

            if watchfiles:
                future = asyncio.create_task(watch())
                return future.cancel
            else:
                logger.warning("watchfiles not installed, cannot watch file")

    solara.use_effect(hot_reload, [value])

    uuid = solara.use_unique_key()
    if css_content_reloaded is not None:
        css_content = css_content_reloaded
    else:
        css_content = value.read_text(encoding="utf-8") if isinstance(value, Path) else value
    # del value
    hash = hashlib.sha256(css_content.encode("utf-8")).hexdigest()
    # the key is unique for this component + value
    # so that we create a new component if the value changes
    # but we do not remove the css of a component with the same value
    key = uuid + "-" + hash
    # ipyvue does not remove the css itself, so we need to do it manually
    script = (
        """
<script>
module.exports = {
  destroyed() {
    document.getElementById("ipyvue-%s").remove();
  },
};
</script>
    """
        % key
    )

    template = f"""
<template>
    <span style="display: none">
    </span>
</template>
{script}
<style id="{key}">
{css_content}
</style>
    """
    # using .key avoids reusing the template, which causes a flicker (due to ipyvue)
    return v.VuetifyTemplate.element(template=template).key(key)



"""
# use_previous

```python
def use_previous(value: T) -> T:
    ...
```

Returns the value from a previous render phase, or the current value on the first render.


"""

import solara

title = "use_previous"


@solara.component
def Page():
    value, set_value = solara.use_state(4)
    value_previous = solara.use_previous(value)
    solara.IntSlider("value", value=value, on_value=set_value)
    solara.Markdown(
        f"""
    **Current**:  `{value}`

    **Previous**: `{value_previous}`
"""
    )



from typing import List, Optional

import solara


def test_toggle_buttons_single():
    value: solara.Reactive[Optional[str]] = solara.reactive(None)

    @solara.component
    def Test():
        with solara.ToggleButtonsSingle("noot", on_value=value.set):
            solara.Button("Aap", value="aap")
            solara.Button("Noot", value="noot")
            solara.Button("Mies", value="mies")
            solara.Button("Nobody", value=None)

    group, rc = solara.render_fixed(Test(), handle_error=False)
    assert group.v_model == 1
    group.v_model = 2
    assert value.value == "mies"
    group.v_model = 3
    assert value.value is None
    # we don't want it to change the index to None
    assert group.v_model == 3


def test_toggle_buttons_multiple():
    value: solara.Reactive[Optional[List[str]]] = solara.reactive(None)

    @solara.component
    def Test():
        with solara.ToggleButtonsMultiple(["noot"], on_value=value.set):
            solara.Button("Aap", value="aap")
            solara.Button("Noot", value="noot")
            solara.Button("Mies", value="mies")

    group, rc = solara.render_fixed(Test())
    assert group.v_model == [1]
    group.v_model = [0, 2]
    assert value.value is not None
    assert value.value == ["aap", "mies"]



from .docs import CategoryLayout, Gallery, NoPage, SubCategoryLayout, WithCode  # noqa
from .header import Header
from .markdown import MarkdownWithMetadata

__all__ = ["Header", "MarkdownWithMetadata", "CategoryLayout", "Gallery", "NoPage", "SubCategoryLayout", "WithCode"]



from typing import Any, Callable, Optional, TypeVar, Union, cast, overload, List, Dict

import ipyvue
import ipyvuetify as vw
import reacton
from typing_extensions import Literal

import solara
from solara.alias import rv as v

T = TypeVar("T")


def use_change(el: reacton.core.Element, on_value: Callable[[Any], Any], enabled=True, update_events=["focusout", "keyup.enter"]):
    """Trigger a callback when a blur events occurs or the enter key is pressed."""
    on_value_ref = solara.use_ref(on_value)
    on_value_ref.current = on_value

    def add_events():
        def on_change(widget, event, data):
            if enabled:
                on_value_ref.current(widget.v_model)

        widget = cast(ipyvue.VueWidget, solara.get_widget(el))
        if enabled:
            for event in update_events:
                widget.on_event(event, on_change)

        def cleanup():
            if enabled:
                for event in update_events:
                    widget.on_event(event, on_change, remove=True)

        return cleanup

    solara.use_effect(add_events, [enabled])


@overload
@solara.component
def InputText(
    label: str,
    value: Union[str, solara.Reactive[str]] = "",
    on_value: Optional[Callable[[str], None]] = ...,
    disabled: bool = ...,
    password: bool = ...,
    continuous_update: bool = ...,
    update_events: list = ...,
    error: Union[bool, str] = ...,
    message: Optional[str] = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
    autofocus: bool = ...,
    dense: bool = ...,
    hide_details: Union[str, bool] = ...,
    placeholder: Optional[str] = ...,
    prefix: Optional[str] = ...,
    suffix: Optional[str] = ...,
    clearable: Literal[False] = ...,
    optional: Literal[False] = ...,
) -> reacton.core.ValueElement[vw.TextField, Any]: ...


@overload
@solara.component
def InputText(
    label: str,
    value: Union[Optional[str], solara.Reactive[Optional[str]]] = ...,
    on_value: Optional[Callable[[Optional[str]], None]] = ...,
    disabled: bool = ...,
    password: bool = ...,
    continuous_update: bool = ...,
    update_events: list = ...,
    error: Union[bool, str] = ...,
    message: Optional[str] = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
    autofocus: bool = ...,
    dense: bool = ...,
    hide_details: Union[str, bool] = ...,
    placeholder: Optional[str] = ...,
    prefix: Optional[str] = ...,
    suffix: Optional[str] = ...,
    clearable: bool = ...,
    optional: Literal[True] = ...,
) -> reacton.core.ValueElement[vw.TextField, Any]: ...


@solara.component
def InputText(
    label: str,
    value: Union[None, str, solara.Reactive[str], solara.Reactive[Optional[str]]] = "",
    on_value: Union[None, Callable[[Optional[str]], None], Callable[[str], None]] = None,
    disabled: bool = False,
    password: bool = False,
    continuous_update: bool = False,
    update_events: List[str] = ["focusout", "keyup.enter"],
    error: Union[bool, str] = False,
    message: Optional[str] = None,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    clearable: bool = False,
    optional: bool = False,
) -> reacton.core.ValueElement[vw.TextField, Optional[str]]:
    """Free form text input.

    ### Basic example:

    ```solara
    import solara

    text = solara.reactive("Hello world!")
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputText("Enter some text", value=text, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: text.set(""))
            solara.Button("Reset", on_click=lambda: text.set("Hello world"))
        solara.Markdown(f"**You entered**: {text.value}")
    ```

    ### Password input:

    This will not show the entered text.

    ```solara
    import solara

    password = solara.reactive("Super secret")
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputText("Enter a password", value=password, continuous_update=continuous_update.value, password=True)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: password.set(""))
            solara.Button("Reset", on_click=lambda: password.set("Super secret"))
        solara.Markdown(f"**You entered**: {password.value}")
    ```


    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `password`: Whether the input is a password input (typically shows input text obscured with an asterisk).
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `update_events`: A list of events that should trigger `on_value`. If continuous update is enabled, this will effectively be ignored,
        since updates will happen every change.
    * `error`: If truthy, show the input as having an error (in red). If a string is passed, it will be shown as the error message.
    * `message`: Message to show below the input. If `error` is a string, this will be ignored.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    * `autofocus`: Determines if a component is to be autofocused or not (Default is False). Autofocus will occur during page load and only one component per page can have autofocus active.
    * `dense`: Reduces the input height.
    * `hide_details`: Hides hint and validation errors. When set to 'auto', messages will be rendered only if there's a message (hint, error message, etc) to display.
    * `placeholder`: Sets the input's placeholder text.
    * `prefix`: Displays prefix text.
    * `suffix`: Displays suffix text.
    * `clearable`: Whether the input can be cleared.
    * `optional`: Whether the value can be None. If `clearable=True`, `optional` is forced to True.
    """
    reactive_value = solara.use_reactive(value, on_value)  # type: ignore
    del value, on_value
    style_flat = solara.util._flatten_style(style)
    classes_flat = solara.util._combine_classes(classes)

    def set_value_cast(value):
        reactive_value.value = str(value)

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    messages = []
    if error and isinstance(error, str):
        messages.append(error)
    elif message:
        messages.append(message)
    text_field = v.TextField(
        v_model=reactive_value.value,
        on_v_model=on_v_model,
        label=label,
        disabled=disabled,
        type="password" if password else "text",
        error=bool(error),
        messages=messages,
        class_=classes_flat,
        style_=style_flat,
        autofocus=autofocus,
        dense=dense,
        hide_details=hide_details,
        placeholder=placeholder if placeholder is not None else "",
        prefix=prefix if prefix is not None else "",
        suffix=suffix if suffix is not None else "",
        clearable=clearable,
    )
    use_change(text_field, set_value_cast, enabled=not continuous_update, update_events=update_events)
    return text_field


@overload
@solara.component
def InputFloat(
    label: str,
    value: Union[float, solara.Reactive[float]] = 0,
    on_value: Optional[Callable[[float], None]] = ...,
    disabled: bool = ...,
    optional: Literal[False] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> reacton.core.ValueElement[vw.TextField, Any]: ...


@overload
@solara.component
def InputFloat(
    label: str,
    value: Union[Optional[float], solara.Reactive[Optional[float]]] = 0,
    on_value: Optional[Callable[[Optional[float]], None]] = ...,
    disabled: bool = ...,
    optional: Literal[True] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> reacton.core.ValueElement[vw.TextField, Any]: ...


@solara.component
def InputFloat(
    label: str,
    value: Union[None, float, solara.Reactive[float], solara.Reactive[Optional[float]]] = 0,
    on_value: Union[None, Callable[[Optional[float]], None], Callable[[float], None]] = None,
    disabled: bool = False,
    optional: bool = False,
    continuous_update: bool = False,
    clearable: bool = False,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
):
    """Numeric input (floats).

    Basic example:

    ```solara
    import solara

    float_value = solara.reactive(42.0)
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputFloat("Enter a float number", value=float_value, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: float_value.set(42.0))
        solara.Markdown(f"**You entered**: {float_value.value}")
    ```


    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `optional`: Whether the value can be None.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `clearable`: Whether the input can be cleared.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    * `autofocus`: Determines if a component is to be autofocused or not (Default is False). Autofocus will occur either during page load, or when the component becomes visible (for example, dialog being opened). Only one component per page should have autofocus on each such event.
    * `dense`: Reduces the input height.
    * `hide_details`: Hides hint and validation errors. When set to 'auto', messages will be rendered only if there's a message (hint, error message, etc) to display.
    * `placeholder`: Sets the input's placeholder text.
    * `prefix`: Displays prefix text.
    * `suffix`: Displays suffix text.

    """

    def str_to_float(value: Optional[str]):
        if value:
            try:
                value = value.replace(",", ".")
                return float(value)
            except ValueError:
                raise ValueError("Value must be a number")
        else:
            if optional:
                return None
            else:
                raise ValueError("Value cannot be empty")

    return _InputNumeric(
        str_to_float,
        label=label,
        value=value,  # type: ignore
        on_value=on_value,  # type: ignore
        disabled=disabled,
        continuous_update=continuous_update,
        clearable=clearable,
        classes=classes,
        style=style,
        autofocus=autofocus,
        dense=dense,
        hide_details=hide_details,
        placeholder=placeholder,
        prefix=prefix,
        suffix=suffix,
    )  # type: ignore


@overload
@solara.component
def InputInt(
    label: str,
    value: Union[int, solara.Reactive[int]] = 0,
    on_value: Optional[Callable[[int], None]] = ...,
    disabled: bool = ...,
    optional: Literal[False] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> reacton.core.ValueElement[vw.TextField, Any]: ...


@overload
@solara.component
def InputInt(
    label: str,
    value: Union[Optional[int], solara.Reactive[Optional[int]]] = 0,
    on_value: Optional[Callable[[Optional[int]], None]] = ...,
    disabled: bool = ...,
    optional: Literal[True] = ...,
    continuous_update: bool = ...,
    clearable: bool = ...,
    classes: List[str] = ...,
    style: Optional[Union[str, Dict[str, str]]] = ...,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> reacton.core.ValueElement[vw.TextField, Any]: ...


@solara.component
def InputInt(
    label: str,
    value: Union[None, int, solara.Reactive[int], solara.Reactive[Optional[int]]] = 0,
    on_value: Union[None, Callable[[Optional[int]], None], Callable[[int], None]] = None,
    disabled: bool = False,
    optional: bool = False,
    continuous_update: bool = False,
    clearable: bool = False,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
):
    """Numeric input (integers).

    Basic example:

    ```solara
    import solara

    int_value = solara.reactive(42)
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputInt("Enter an integer number", value=int_value, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: int_value.set(42))
        solara.Markdown(f"**You entered**: {int_value.value}")
    ```

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `optional`: Whether the value can be None.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `clearable`: Whether the input can be cleared.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    * `autofocus`: Determines if a component is to be autofocused or not (Default is False). Autofocus will occur either during page load, or when the component becomes visible (for example, dialog being opened). Only one component per page should have autofocus on each such event.
    * `dense`: Reduces the input height.
    * `hide_details`: Hides hint and validation errors. When set to 'auto', messages will be rendered only if there's a message (hint, error message, etc) to display.
    * `placeholder`: Sets the input's placeholder text.
    * `prefix`: Displays prefix text.
    * `suffix`: Displays suffix text.
    """

    def str_to_int(value: Optional[str]):
        if value:
            try:
                return int(value)
            except ValueError:
                raise ValueError("Value must be an integer")
        else:
            if optional:
                return None
            else:
                raise ValueError("Value cannot be empty")

    return _InputNumeric(
        str_to_int,
        label=label,
        value=value,  # type: ignore
        on_value=on_value,  # type: ignore
        disabled=disabled,
        continuous_update=continuous_update,
        clearable=clearable,
        classes=classes,
        style=style,
        autofocus=autofocus,
        dense=dense,
        hide_details=hide_details,
        placeholder=placeholder,
        prefix=prefix,
        suffix=suffix,
    )  # type: ignore


def _use_input_type(
    input_value: Union[None, T, solara.Reactive[Optional[T]], solara.Reactive[T]],
    parse: Callable[[Optional[str]], T],
    stringify: Callable[[Optional[T]], str],
    on_value: Union[None, Callable[[Optional[T]], None], Callable[[T], None]] = None,
):
    reactive_value = solara.use_reactive(input_value, on_value)  # type: ignore
    del input_value, on_value
    string_value, set_string_value = solara.use_state(stringify(reactive_value.value) if reactive_value.value is not None else None)  # type: ignore
    # Use a ref to make sure sync_back_input_value() does not get a stale string_value
    string_value_ref = solara.use_ref(string_value)
    string_value_ref.current = string_value

    error_message = cast(Union[str, None], None)

    try:
        reactive_value.set(parse(string_value))  # type: ignore
    except ValueError as e:
        error_message = str(e.args[0])

    def sync_back_input_value():
        # Make sure we update string_value when the effect is rerun,
        # Since the parsing & stringigying functions might have changed
        set_string_value(stringify(reactive_value.value) if reactive_value.value is not None else None)  # type: ignore

        def on_external_value_change(new_value: Optional[T]):
            new_string_value = stringify(new_value)
            try:
                parse(string_value_ref.current)
            except ValueError:
                # String value could be invalid when external value is changed by a different component
                set_string_value(new_string_value)
            else:
                if new_value != parse(string_value_ref.current):
                    set_string_value(new_string_value)

        return reactive_value.subscribe(on_external_value_change)  # type: ignore

    solara.use_effect(sync_back_input_value, [reactive_value, parse, stringify])

    return string_value, error_message, set_string_value


@solara.component
def _InputNumeric(
    str_to_numeric: Callable[[Optional[str]], T],
    label: str,
    value: Union[None, T, solara.Reactive[Optional[T]], solara.Reactive[T]],
    on_value: Union[None, Callable[[Optional[T]], None], Callable[[T], None]] = None,
    disabled: bool = False,
    continuous_update: bool = False,
    clearable: bool = False,
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
    autofocus: bool = False,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
):
    """Numeric input.

    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    * `autofocus`: Determines if a component is to be autofocused or not (Default is False). Autofocus will occur either during page load, or when the component becomes visible (for example, dialog being opened). Only one component per page should have autofocus on each such event.
    * `dense`: Reduces the input height.
    * `hide_details`: Hides hint and validation errors. When set to 'auto', messages will be rendered only if there's a message (hint, error message, etc) to display.
    * `placeholder`: Sets the input's placeholder text.
    * `prefix`: Displays prefix text.
    * `suffix`: Displays suffix text.
    * `clearable`: Whether the input can be cleared.
    """
    style_flat = solara.util._flatten_style(style)
    classes_flat = solara.util._combine_classes(classes)

    internal_value, error, set_value_cast = _use_input_type(
        value,
        str_to_numeric,
        str,
        on_value,
    )

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    if error:
        label += f" ({error})"
    text_field = v.TextField(
        v_model=internal_value,
        on_v_model=on_v_model,
        label=label,
        disabled=disabled,
        # we are not using the number type, since we cannot validate invalid input
        # see https://stackoverflow.blog/2022/12/26/why-the-number-input-is-the-worst-input/
        # type="number",
        hide_details=hide_details,
        clearable=clearable,
        error=bool(error),
        class_=classes_flat,
        style_=style_flat,
        autofocus=autofocus,
        dense=dense,
        placeholder=placeholder if placeholder is not None else "",
        prefix=prefix if prefix is not None else "",
        suffix=suffix if suffix is not None else "",
    )
    use_change(text_field, set_value_cast, enabled=not continuous_update)
    return text_field



import solara
from .markdown import MarkdownWithMetadata
from .breadcrumbs import BreadCrumbs


@solara.component
def Gallery(route_external=None):
    from ..pages.documentation.examples import pycafe_projects

    if route_external is not None:
        route_current = route_external
    else:
        # show a gallery of all the examples
        router = solara.use_router()
        route_current = router.path_routes[-2]

    with solara.Column(gap="75px", align="center"):
        with solara.Row(justify="center", gap="30px", style={"margin": "30px 0 !important", "flex-wrap": "wrap", "align-items": "start", "row-gap": "30px"}):
            for child in route_current.children:
                if child.path == "/":
                    continue
                with solara.v.Html(tag="a", attributes={"href": "#" + child.path}):
                    solara.Button(
                        child.label,
                        color="primary",
                        classes=["v-btn--rounded", "v-size--x-large"],
                    )

        with solara.Column(gap="75px", style={"padding-bottom": "75px"}):
            for route in route_current.children:
                if route.children:
                    with solara.Column(classes=["subcategory-row", "ps-md-10"]):
                        solara.HTML(tag="h2", unsafe_innerHTML=route.label, attributes={"id": route.path}, style="padding-left: 10%;")
                        with solara.Row(justify="center", gap="20px", style={"flex-wrap": "wrap", "row-gap": "20px", "max-width": "90rem"}):
                            for child in route.children:
                                if child.path == "/":
                                    continue
                                path = route.path + "/" + child.path
                                if child.path in [
                                    "button",
                                    "checkbox",
                                    "confirmation_dialog",
                                    "echarts",
                                    "file_browser",
                                    "file_download",
                                    "matplotlib",
                                    "select",
                                    "switch",
                                    "tooltip",
                                ]:
                                    image_url = "https://dxhl76zpt6fap.cloudfront.net/public/api/" + child.path + ".gif"
                                elif child.path in ["card", "dataframe", "pivot_table", "slider"]:
                                    image_url = "https://dxhl76zpt6fap.cloudfront.net/public/api/" + child.path + ".png"
                                elif child.path in pycafe_projects:
                                    image_url = f"https://py.cafe/preview/solara/{child.path}"
                                else:
                                    image_url = "https://dxhl76zpt6fap.cloudfront.net/public/logo.svg"

                                if path:
                                    path = path if route_external is None else route_current.path + "/" + path
                                    title = solara.Link(path, children=[child.label])
                                    with solara.Card(title, classes=["component-card"], margin=0):
                                        with solara.Column(align="center"):
                                            with solara.Link(path):
                                                solara.Image(image_url, width="12rem")


@solara.component
def NoPage():
    raise RuntimeError("This page should not be rendered")


@solara.component
def WithCode(route_current):
    component = getattr(route_current.module, "Page", None)
    with solara.Column(style={"flex-grow": 1, "padding-top": "56px"}) as main:
        BreadCrumbs()
        # It renders code better
        MarkdownWithMetadata(
            route_current.module.__doc__ or "# no docs yet",
            unsafe_solara_execute=True,
        )
        if component and component != NoPage:
            with solara.Card("Example", margin=0, classes=["mt-8"]):
                component()
                github_url = solara.util.github_url(route_current.module.__file__)
                solara.Button(
                    label="View source",
                    icon_name="mdi-github-circle",
                    attributes={"href": github_url, "target": "_blank"},
                    text=True,
                    outlined=True,
                    class_="mt-8",
                )
    return main


@solara.component
def SubCategoryLayout(children=[]):
    route_current, all_routes = solara.use_route()
    router = solara.use_router()
    sibling_routes = router.path_routes[-2]
    if route_current is None:
        return solara.Error("Page not found")
    elif route_current.path == "/":
        with solara.Column(
            gap="10px", classes=["docs-card-container"], style={"flex-grow": 1, "max-width": "80%", "width": "550px", "padding-top": "64px"}, align="stretch"
        ):
            solara.HTML(tag="h2", unsafe_innerHTML=route_current.label, attributes={"id": route_current.path}, style="padding-left: 10%;")
            for route in sibling_routes.children:
                if route.path == "/":
                    continue
                with solara.Link(route.path):
                    with solara.Row(
                        classes=["docs-card"],
                        style={
                            "background-color": "var(--docs-color-grey)",
                            "align-items": "center",
                            "height": "3rem",
                        },
                    ):
                        solara.HTML(tag="h3", unsafe_innerHTML=route.label, style={"color": "white", "display": "block", "flex-grow": "1", "padding": "0 24px"})
                        solara.v.Icon(children=["mdi-arrow-right"], color="var(--color-grey-light)", class_="docs-card-icon")
    elif route_current.module:
        WithCode(route_current)
    else:
        with solara.Column(align="center", children=children, style={"flex-grow": 1, "padding": "0"}) as main:
            pass
        return main


@solara.component
def CategoryLayout(children=[]):
    route_current, all_routes = solara.use_route()
    if route_current is None:
        return solara.Error("Page not found")

    if route_current.path == "/":
        return Gallery()
    else:
        with solara.Column(align="stretch", children=children, style={"max-width": "100%"}) as main:
            pass
        return main



import solara
import solara.lab
import ipyvuetify as v
import time


def test_docs_no_browser_api_thread():
    clicks = solara.reactive(0)

    @solara.component
    def ClickButton():
        @solara.lab.task
        def increment():
            # now we will wait for 0.3 seconds before updating the UI
            time.sleep(0.3)
            clicks.value += 1

        with solara.Card("Button in a card"):
            with solara.Column():
                solara.Button(label=f"Clicked: {clicks}", on_click=increment)

    # rc is short for render context
    box, rc = solara.render(ClickButton(), handle_error=False)
    finder = rc.find(v.Btn)
    button = finder.widget
    finder.assert_single()
    finder.assert_not_empty()
    assert button.children[0] == "Clicked: 0"

    # clicking will now start a thread, so we have to wait/poll for the UI to update
    button.click()

    button_after_delayed_click = rc.find(v.Btn, children=["Clicked: 1"])
    button_after_delayed_click.wait_for(timeout=2.5)



from pathlib import Path

from solara.website.components.sidebar import Sidebar
from solara.website.components import MarkdownWithMetadata

title = "Changelog"
HERE = Path(__file__)

Page = MarkdownWithMetadata(Path(HERE.parent / "changelog.md").read_text())
Sidebar = Sidebar



"""# Pokemon search

This example shows to to dynamically fetch data and render the results in a table when done, and show an error on failure.

It also shows an optional filter to narrow down the results.
"""

import solara
from solara import use_fetch
from solara.alias import rv

github_url = solara.util.github_url(__file__)
pokemon_base_url = "https://raw.githubusercontent.com/jherr/pokemon/0722479d4153b1db0d0326956b08b37f44a95a5f"
json_url = f"{pokemon_base_url}/index.json"


@solara.component
def Page():
    data = use_fetch(json_url)
    json = solara.use_json_load(data)
    filter, set_filter = solara.use_state("")

    with solara.Div() as main:
        if json.error:
            solara.Error(f"Error {json.error}")
            solara.Button("Retry", on_click=data.retry)
        else:
            if json.value:
                solara.InputText(label="Filter pokemons by name", value=filter, on_value=set_filter, continuous_update=True)
                pokemons = json.value
                if filter:
                    pokemons = [k for k in pokemons if filter.lower() in k["name"].lower()]
                    if len(pokemons) == 0:
                        solara.Warning("No pokemons found, try a different filter")
                    else:
                        solara.Info(f"{len(pokemons)} pokemons found")
                else:
                    solara.Info(f"{len(pokemons)} pokemons in total")
                with solara.GridFixed(columns=4, align_items="end", justify_items="stretch"):
                    for pokemon in pokemons[:20]:
                        with solara.Div():
                            name = pokemon["name"]
                            url = f"{pokemon_base_url}/{pokemon['image']}"
                            # TODO: how to do this with solara
                            rv.Img(src=url, contain=True, max_height="200px")
                            solara.Text(name)
            else:
                with solara.Div():
                    solara.Text("Loading...")
                    rv.ProgressCircular(indeterminate=True, class_="solara-progress")
    return main



"""
# AppLayout

"""

import solara
import solara.lab
from solara.website.utils import apidoc

title = "AppLayout"


@solara.component
def Page():
    return solara.Markdown(
        """
    An example cannot be shown embedded in this page, Visit the [AppLayout page](/apps/scatter) to see an example.

    [![AppLayout screenshot](https://dxhl76zpt6fap.cloudfront.net/public/docs/app-layout.webp)](/apps/scatter)
    """
    )


__doc__ += apidoc(solara.AppLayout.f)  # type: ignore



from pathlib import Path
from typing import List
import requests


import solara.server.starlette
import solara.server.settings
import contextlib

import playwright.sync_api
from playwright.sync_api import expect


HERE = Path(__file__).parent


@contextlib.contextmanager
def extra_assets(locations: List[str]):
    prev = solara.server.settings.assets.extra_locations
    solara.server.settings.assets.extra_locations = locations
    try:
        yield
    finally:
        solara.server.settings.assets.extra_locations = prev


def test_assets_extra(solara_server):
    response = requests.get(f"{solara_server.base_url}/static/assets/common.js")
    assert response.status_code == 404

    with extra_assets([str(HERE / "assets" / "assets1")]):
        response = requests.get(f"{solara_server.base_url}/static/assets/common.js")
        assert response.status_code == 200
        assert response.text.strip() == "content1"
        response = requests.get(f"{solara_server.base_url}/static/assets/unique1.js")
        assert response.status_code == 200
        response = requests.get(f"{solara_server.base_url}/static/assets/unique2.js")
        assert response.status_code == 404

        response = requests.get(f"{solara_server.base_url}/static/assets/custom.js")
        assert response.status_code == 200
        assert response.text.strip() == "/* not empty */"

    with extra_assets([str(HERE / "assets" / "assets2")]):
        response = requests.get(f"{solara_server.base_url}/static/assets/common.js")
        assert response.status_code == 200
        assert response.text.strip() == "content2"
        response = requests.get(f"{solara_server.base_url}/static/assets/unique1.js")
        assert response.status_code == 404
        response = requests.get(f"{solara_server.base_url}/static/assets/unique2.js")
        assert response.status_code == 200

        response = requests.get(f"{solara_server.base_url}/static/assets/custom.js")
        assert response.status_code == 200
        assert response.text.strip() == "var a = 1;"

    with extra_assets([str(HERE / "assets" / "assets1"), str(HERE / "assets" / "assets2")]):
        response = requests.get(f"{solara_server.base_url}/static/assets/common.js")
        assert response.status_code == 200
        assert response.text.strip() == "content1"
        response = requests.get(f"{solara_server.base_url}/static/assets/unique1.js")
        assert response.status_code == 200
        response = requests.get(f"{solara_server.base_url}/static/assets/unique2.js")
        assert response.status_code == 200


def test_api_style(page_session: playwright.sync_api.Page, solara_server, solara_app):
    # this test is added because the include css macro in the jinja template may embed the
    # css, which also requires respecting the extra assets directories
    with solara_app("solara.website.pages.documentation.components.input.button"), extra_assets(
        [str(HERE / "assets" / "assets1"), str(HERE / "assets" / "assets2")]
    ):
        page_session.goto(solara_server.base_url)
        button = page_session.locator("text=Clicked 0 times")
        button.first.wait_for()
        expect(button).to_have_css("color", "rgb(255, 0, 0)")



import dataclasses
from pathlib import Path
from typing import Any, Dict

import vaex.datasets
import yaml


@dataclasses.dataclass
class DataFrame:
    title: str
    df: Any
    image_url: str


dfs = {
    "titanic": DataFrame(
        df=vaex.datasets.titanic(),
        title="Titanic",
        image_url="https://images.unsplash.com/photo-1561625116-df74735458a5?ixlib=rb-1.2.1&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=3574&q=80",  # noqa
    ),
    "iris": DataFrame(
        df=vaex.datasets.iris(),
        title="Iris",
        image_url="https://images.unsplash.com/photo-1540163502599-a3284e17072d?ixlib=rb-1.2.1&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=3870&q=80",  # noqa
    ),
    # uncomment for a larger dataset to be included
    # "taxi": DataFrame(
    #     df=vaex.datasets.taxi(),
    #     title="New York Taxi",
    #     image_url="https://images.unsplash.com/photo-1514749204155-24e484635226?ixlib=rb-1.2.1&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=1587&q=80",  # noqa
    # ),
}

names = list(dfs)
# def load(name):
#     if name == "titanic"

HERE = Path(__file__)


@dataclasses.dataclass
class Article:
    markdown: str
    title: str
    description: str
    image_url: str


articles: Dict[str, Article] = {}

for file in (HERE.parent / "content/articles").glob("*.md"):
    content = file.read_text()
    lines = [k.strip() for k in content.split("\n")]
    frontmatter_start = lines.index("---", 0)
    frontmatter_end = lines.index("---", frontmatter_start + 1)
    yamltext = "\n".join(lines[frontmatter_start + 1 : frontmatter_end - 2])
    metadata = yaml.safe_load(yamltext)
    markdown = "\n".join(lines[frontmatter_end + 1 :])
    articles[file.stem] = Article(markdown=markdown, title=metadata["title"], description=metadata["description"], image_url=metadata["image"])



"""
# Chatbot

A way to create a chatbot using OpenAI's GPT-4 API, utilizing their new API, and the streaming feature.
"""

import os
from typing import List, cast

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionMessageParam
from typing_extensions import TypedDict

import solara
import solara.lab


class MessageDict(TypedDict):
    role: str  # "user" or "assistant"
    content: str


messages: solara.Reactive[List[MessageDict]] = solara.reactive([])

try:
    import pycafe

    OPENAI_API_KEY = pycafe.get_secret(
        "OPENAI_API_KEY",
        """We need an OpenAI API key to generate text.

Go to [OpenAI](https://platform.openai.com/account/api-keys) to get one.

Or read [this](https://www.rebelmouse.com/openai-account-set-up) article for
more information.

Or read more [about secrets on PyCafe](/docs/secrets)

""",
    )
except ModuleNotFoundError:
    OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

openai = AsyncOpenAI(api_key=OPENAI_API_KEY) if OPENAI_API_KEY else None


def no_api_key_message():
    messages.value = [
        {
            "role": "assistant",
            "content": "No OpenAI API key found. Please set your OpenAI API key in the environment variable `OPENAI_API_KEY`.",
        },
    ]


@solara.lab.task
async def promt_ai(message: str):
    if openai is None:
        no_api_key_message()
        return

    messages.value = [
        *messages.value,
        {"role": "user", "content": message},
    ]
    # The part below can be replaced with a call to your own
    response = await openai.chat.completions.create(
        model="gpt-4-1106-preview",
        # our MessageDict is compatible with the OpenAI types
        messages=cast(List[ChatCompletionMessageParam], messages.value),
        stream=True,
    )
    # start with an empty reply message, so we render and empty message in the chat
    # while the AI is thinking
    messages.value = [*messages.value, {"role": "assistant", "content": ""}]
    # and update it with the response
    async for chunk in response:
        if chunk.choices[0].finish_reason == "stop":  # type: ignore
            return
        # replace the last message element with the appended content
        delta = chunk.choices[0].delta.content
        assert delta is not None
        updated_message: MessageDict = {
            "role": "assistant",
            "content": messages.value[-1]["content"] + delta,
        }
        # replace the last message element with the appended content
        # which will update the UI
        messages.value = [*messages.value[:-1], updated_message]


@solara.component
def Page():
    with solara.Column(
        style={"width": "100%", "height": "50vh"},
    ):
        with solara.lab.ChatBox():
            for item in messages.value:
                with solara.lab.ChatMessage(
                    user=item["role"] == "user",
                    avatar=False,
                    name="ChatGPT" if item["role"] == "assistant" else "User",
                    color="rgba(0,0,0, 0.06)" if item["role"] == "assistant" else "#ff991f",
                    avatar_background_color="primary" if item["role"] == "assistant" else None,
                    border_radius="20px",
                ):
                    solara.Markdown(item["content"])
        if promt_ai.pending:
            solara.Text("I'm thinking...", style={"font-size": "1rem", "padding-left": "20px"})
            solara.ProgressLinear()
        # if we don't call .key(..) with a unique key, the ChatInput component will be re-created
        # and we'll lose what we typed.
        solara.lab.ChatInput(send_callback=promt_ai, disabled_send=promt_ai.pending, autofocus=True).key("input")



import solara
from solara.website.components.contact import Contact


@solara.component
def Page():
    with solara.Column(gap="40px", align="stretch", style={"width": "100%", "padding-bottom": "140px"}):
        with solara.Div(style={"display": "none"}):
            title = "Scaling up your ipywidgets apps"
            solara.Meta(property="og:title", content=title)
            solara.Meta(name="twitter:title", content=title)
            solara.Title(title)

            description = "Learn how to transform your ipywidget app or prototype into a production ready high quality app."
            solara.Meta(name="description", property="og:description", content=description)
            solara.Meta(name="twitter:description", content=description)

        solara.Markdown("""
# Scaling up your ipywidgets app

## The problem
We at Widgetti, the creators of Solara, have regularly seen companies struggle trying to scale up their ipywidgets apps.

These app often start as a prototype or a small app in the Jupyter notebook, and become important internal apps. Together with Voila
it allows companies to quickly deploy an app internally. However, as the app grows, the complexity of the app grows as well and the app
becomes harder to maintain and scale, and extend and testing is often an afterthought.

## Our solution
If you are in this position, you are not alone. We have seen this happen many times and we have helped many companies to scale up their ipywidgets apps.
This can go from small improvements to the app, to a complete rewrite of the app. We often recomment an incremental approach, where we make sure the app
gets tested properly, and gradually introduce best practices to the app and your team as well as introducing modern state management techniques.

We continuously do code review and assist your team in making the app more maintainable and scalable. We can also help with the deployment of the app.

Over time, we phase out our involvement and make sure your team is able to maintain and extend the app themselves. We can also provide training and workshops
if needed.

Please contact us below for a free consultation where we can discuss your needs and how we can help you, or send us an email at [contact@solara.dev](mailto:contact@solara.dev).

                        """)
        Contact(
            title="Get In Touch",
            subtitle="Please fill out the form below and we will get back to you as soon as possible.",
            email_subject="Scaling up your ipywidgets app",
        )



import solara

title = "Showcase"


@solara.component
def Page():
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Card("Wanderlust", style={"height": "100%"}):
            solara.Markdown(
                """
            [Wanderlust](/showcase/wanderlust) is a reproduction of the travel assistant demo shown at the
            [OpenAI DevDay](https://devday.openai.com/) 2023, built using Solara and the OpenAI Assistants API.
            """
            )
            with solara.Link("./wanderlust"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/wanderlust/thumbnail.png", width="100%")
            with solara.v.Html(tag="a", attributes={"href": "https://github.com/widgetti/wanderlust", "target": "_blank"}):
                with solara.Row(style={"min-height": "24px"}):
                    solara.v.Icon(children=["mdi-github-circle"], x_large=True, class_="mr-2")

        with solara.Card("Domino Code Assist", style={"height": "100%"}):
            solara.Markdown(
                """
            Domino Code Assist (DCA) is a tool developed for Domino Data Lab which provides a simple, intuitive point-and-click interface for generating
            Python or R code.
            """
            )
            with solara.Link("./domino_code_assist"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/lca/thumbnail.png", width="100%")

        with solara.Card("TESSA", style={"height": "100%"}):
            solara.Markdown(
                """
            [TESSA](https://planeto-energy.ch/) is a tool developed by Planeto for district heating & cooling planning.
            """
            )
            with solara.Link("./planeto_tessa"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/tessa/thumbnail.png", width="100%")

        with solara.Card("Bulk labeling", style={"height": "100%"}):
            solara.Markdown(
                """
            [Bulk labeling](https://github.com/Ben-Epstein/solara-examples/tree/main/bulk-labeling)
            is a tool developed by Ben Epstein for labeling data in bulk."""
            )
            from solara.alias import rv

            img = "https://avatars.githubusercontent.com/u/22605641?v=4"
            with rv.Html(tag="a", attributes={"href": "https://github.com/Ben-Epstein/solara-examples/tree/main/bulk-labeling", "target": "_blank"}):
                solara.Image("https://user-images.githubusercontent.com/1765949/237517090-8f7242b1-3189-4c5b-abd3-0f0986292ade.png", width="100%")
            with rv.Html(tag="a", attributes={"href": "https://github.com/Ben-Epstein/", "target": "_blank"}):
                with rv.ListItem(class_="grow"):
                    with rv.ListItemAvatar(color="grey darken-3"):
                        rv.Img(
                            class_="elevation-6",
                            src=img,
                        )

        with solara.Card("solara.dev", style={"height": "100%"}):
            solara.Markdown(
                """
            The Solara.dev website is built using Solara itself, showcasing the library's features and capabilities.
            All documentation and example are on this page.
            """
            )
            with solara.Link("./solara_dev"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solara/thumbnail.png", width="100%")

    solara.HTML(tag="h1", unsafe_innerHTML="Solarathon 2023")
    solara.Markdown(
        """
                    In November 2023, we hosted the first Solarathon, a two-week hackathon,
                    during which participants built projects using Solara, and deployed them to [Ploomber](https://ploomber.io/).
                    The event was a great success, and we are looking forward to hosting more Solarathons in the future.
                    The projects built during the event are showcased below."""
    )
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Card("Team 2: Travel Assistant", style={"height": "100%"}):
            with solara.Link("./solarathon_2023_team_2"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_2/thumbnail.png", width="100%")
            with solara.v.Html(tag="a", attributes={"href": "https://github.com/alonsosilvaallende/solarathon", "target": "_blank"}):
                with solara.Row(style={"min-height": "24px"}):
                    solara.v.Icon(children=["mdi-github-circle"], x_large=True, class_="mr-2")

        with solara.Card("Team 4: Live Cryptocurrency Dashboard", style={"height": "100%"}):
            with solara.Link("./solarathon_2023_team_4"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_4/thumbnail.png", width="100%")
            with solara.v.Html(tag="a", attributes={"href": "https://github.com/theeldermillenial/solarathon", "target": "_blank"}):
                with solara.Row(style={"min-height": "24px"}):
                    solara.v.Icon(children=["mdi-github-circle"], x_large=True, class_="mr-2")

        with solara.Card("Team 5: Sports Video Analysis", style={"height": "100%"}):
            with solara.Link("./solarathon_2023_team_5"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_5/thumbnail.png", width="100%")
            with solara.v.Html(tag="a", attributes={"href": "https://github.com/FelipeCabelloE/solarathon-team5/", "target": "_blank"}):
                with solara.Row(style={"min-height": "24px"}):
                    solara.v.Icon(children=["mdi-github-circle"], x_large=True, class_="mr-2")

        with solara.Card("Team 6: Automatic FAQ Creator", style={"height": "100%"}):
            with solara.Link("./solarathon_2023_team_6"):
                solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_6/thumbnail.png", width="100%")
            with solara.v.Html(tag="a", attributes={"href": "https://github.com/silvhua/solarathon-faq-creator", "target": "_blank"}):
                with solara.Row(style={"min-height": "24px"}):
                    solara.v.Icon(children=["mdi-github-circle"], x_large=True, class_="mr-2")
